from dlk.utils.vocab import Vocabulary
from collections import Counter
from typing import Dict, Iterable, List, Union
import pytest

class TestVocabulary(object):
    """generate vocab from tokens(token or Iterable tokens)
       you can dumps the object to dict and load from dict
    """

    @pytest.mark.parametrize('tokens', [
                                 ['b', 'a', 'c'],
                                 ['a', 'c'],
                                 'a'
                             ])
    def test_dumps_and_load(self, tokens):
        """test before and after dump the class is same

        """
        vocab = Vocabulary()
        vocab.auto_update(tokens)
        assert Vocabulary.load(vocab.dumps()).__dict__ == vocab.__dict__

    @pytest.mark.parametrize('tokens', [
                                 ['aa', 'aa', 'ab', 'ba'],
                             ])
    def test_filter_rare(self, tokens):
        """test filter the rare appear words

        """
        vocab = Vocabulary()
        vocab.auto_update(tokens)
        vocab.filter_rare(most_common=1)
        assert vocab.word2idx == {'aa': 0}

        vocab = Vocabulary()
        vocab.auto_update(tokens)
        vocab.filter_rare(min_freq=2)
        assert vocab.word2idx == {'aa': 0}

    @pytest.mark.parametrize('tokens', [
                                 ['aa', 'aa', 'ab', 'ba'],
                             ])
    def test_add_autoadd(self, tokens):
        """test add or autoadd do the same thing
        """
        vocab_a = Vocabulary()
        vocab_a.auto_update(tokens)

        vocab_b = Vocabulary()
        for token in tokens:
            vocab_b.add(token)

        vocab_c = Vocabulary()
        vocab_c.add_from_iter(tokens)
        assert vocab_a.dumps() == vocab_b.dumps() == vocab_c.dumps()



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Callable, Dict

PROTECTED = ["_name", "_base", "_search"]


def register_module_name(module_name: str):
    """get the real registerd instance module name

    Args:
        module_name: the configure module name

    Returns:
        instance module name
    """
    return module_name.split("-")[0]


class Register(object):
    """Register"""

    def __init__(self):
        self.registry: Dict[str, Dict[str, Any]] = {}

    def register(self, type_name: str, name: str) -> Callable:
        """register the name: module to self.registry

        Args:
            name: the registered module name

        Returns:
            the module

        """

        def decorator(module):
            if type_name not in self.registry:
                self.registry[type_name] = {}
            if name.strip() == "":
                raise ValueError(f"You must set a name for {module.__name__}")

            if name in self.registry[type_name]:
                raise ValueError(f"The {name} is already registered in {type_name}.")
            self.registry[type_name][name] = module
            return module

        return decorator

    def __call__(self, type_name: str, name: str) -> Callable:
        """you can directly call the object, the behavior is the same as object.register(name)"""
        return self.register(type_name, name)

    def get(self, type_name: str, name: str) -> Any:
        """get the module by name

        Args:
            name: the name should be the real name or name+@+sub_name, and the

        Returns:
            registered module

        """
        if type_name not in self.registry:
            raise KeyError(f"There is not a registerd type named '{type_name}'")
        if name not in self.registry[type_name]:
            raise KeyError(
                f"In '{type_name}' register, there is not a entry named '{name}'"
            )
        return self.registry[type_name][name]

    def __getitem__(self, type_and_name: tuple) -> Any:
        """wrap for object.get(name)"""
        type_name, name = type_and_name
        return self.get(type_name, name)


register = Register()



# Copyright the author(s) of DLK.

# There are many code copied from fairseq
# Copyright (c) Facebook, Inc. and its affiliates.


import copy
import math
from typing import Dict, List, Optional

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
    dataclass,
)
from tokenizers import Tokenizer

from dlk.nn.base_module import BaseModel
from dlk.utils.ngram_repeat_block import NGramRepeatBlock
from dlk.utils.register import register, register_module_name


@dataclass
class TokenGenBaseConfig(Base):
    beam_size = IntField(value=1, help="beam search size (default: 1)")
    share_embedding = BoolField(value=True, help="share embedding or not")
    embed_scale = StrField(
        value="none",
        options=["none", "default"],
        help="the embed scale, if `none` will not scale the embedding, if `default` will scale the embedding as sqrt(embed_dim)",
    )
    model_type = StrField(
        value=MISSING,
        options=["dec_only", "token_enc_dec", "media_enc_token_dec"],
        help="the model type",
    )

    tgt_tokenizer = StrField(value=MISSING, help="the target tokenizer file")
    tgt_embedding_dim = IntField(value="???", help="the target embedding dim")
    decoder_hidden_size = IntField(
        value=None,
        additions=[None],
        help="the hidden size of the decoder, if set to None, will use the `tgt_embedding_dim`",
    )
    lm_head_bias = BoolField(value=True, help="the lm head bias")
    normalize_scores = BoolField(
        value=True, help="normalize scores by the length of the output (default: True)"
    )
    len_penalty = FloatField(
        value=1.0,
        help="length penalty, where <1.0 favors shorter, >1.0 favors longer sentences (default: 1.0)",
    )
    unk_penalty = FloatField(
        value=0.0,
        help="unknown word penalty, where <0 produces more unks, >0 produces fewer (default: 0.0)",
    )
    temperature = FloatField(
        value=1.0,
        help="temperature, where values >1.0 produce more uniform samples and values <1.0 produce sharper samples (default: 1.0)",
    )
    no_repeat_ngram_size = IntField(
        value=0,
        minimum=1,
        additions=[0],
        help="prevent ngram repeat, only support 0 or >=2",
    )
    min_len = IntField(
        value=1,
        help="the minimum length of the generated output(not including end-of-sentence)",
    )
    tgt_eos = StrField(value=MISSING, help="the end of sentence token")
    tgt_pad = StrField(value=MISSING, help="the padding token")
    tgt_unk = StrField(value=MISSING, help="the unknown token")
    tgt_bos = StrField(
        value=None,
        additions=[None],
        help="the begin of sentence token, if None, the eos token will be used",
    )

    class InputMap:
        encoder_input_ids = StrField(
            value="encoder_input_ids", help="the encode input ids"
        )
        decoder_input_ids = StrField(
            value="decoder_input_ids", help="the encode input ids"
        )

    input_map = NestField(
        value=InputMap, help="the input map of the token_enc_dec module"
    )

    class OutputMap:
        decoder_target_ids = StrField(value="decoder_target_ids", help="the target ids")
        generated = StrField(value="generated", help="the generated output")
        logits = StrField(value="logits", help="the logits of the token_enc_dec module")

    output_map = NestField(
        value=OutputMap, help="the output map of the token_enc_dec module"
    )

    submodule = SubModule(
        value={},
        suggestions=[
            "encoder",
            "token_gen_decoder",
            "initmethod",
            "token_sample",
        ],
        help="submodules for token_enc_dec model",
    )


class SpecialVocab(object):
    """You should not init this class from scratch"""

    def __init__(self, config):
        self.tokenizer = Tokenizer.from_file(config.tgt_tokenizer)

        self._eos = config.tgt_eos
        self._bos = config.tgt_bos
        self._pad = config.tgt_pad
        self._unk = config.tgt_unk

    def eos(self):
        """
        Returns:
            eos value
        """
        return self.tokenizer.token_to_id(self._eos)

    def bos(self):
        """
        Returns:
            bos value
        """
        if self._bos is None:
            return self.tokenizer.token_to_id(self._eos)
        return self.tokenizer.token_to_id(self._bos)

    def pad(self):
        """
        Returns:
            eos value
        """
        return self.tokenizer.token_to_id(self._pad)

    def unk(self):
        """
        Returns:
            eos value
        """
        return self.tokenizer.token_to_id(self._unk)

    def __len__(self):
        """
        Returns:
            tokenizer vocab size

        """
        return self.tokenizer.get_vocab_size()


class TokenGenBase(nn.Module):
    def __init__(
        self, config: TokenGenBaseConfig, checkpoint, encoder_type: str = "encoder"
    ):
        """Generates translations of a given source sentence.
        Args:
        """
        super().__init__()
        self.config = config
        self.tgt_dict = SpecialVocab(config)
        self.tgt_vocab_size = len(self.tgt_dict)
        self.embedding_scale = (
            math.sqrt(config.tgt_embedding_dim)
            if config.embed_scale == "default"
            else 1.0
        )
        if self.config.decoder_hidden_size is None:
            self.config.decoder_hidden_size = self.config.tgt_embedding_dim

        if config.share_embedding:
            self.embedding = nn.Embedding(
                self.tgt_vocab_size, self.config.tgt_embedding_dim
            )
        else:
            self.tgt_embedding = nn.Embedding(
                self.tgt_vocab_size, self.config.tgt_embedding_dim
            )

        encoder_configs = config._get_modules(encoder_type)
        assert len(encoder_configs) == 1
        if config.share_embedding:
            self.encoder = register.get(
                encoder_type,
                register_module_name(encoder_configs[0]._module_name),
            )(config=encoder_configs[0], embedding=self.embedding)
        else:
            self.encoder = register.get(
                encoder_type,
                register_module_name(encoder_configs[0]._module_name),
            )(config=encoder_configs[0], embedding=self.src_embedding)

        decoder_configs = config._get_modules("token_gen_decoder")
        assert len(decoder_configs) == 1
        if config.share_embedding:
            self.decoder = register.get(
                "token_gen_decoder",
                register_module_name(decoder_configs[0]._module_name),
            )(config=decoder_configs[0], embedding=self.embedding)
        else:
            self.decoder = register.get(
                "token_gen_decoder",
                register_module_name(decoder_configs[0]._module_name),
            )(config=decoder_configs[0], embedding=self.tgt_embedding)

        self.lm_head = nn.Linear(
            self.config.decoder_hidden_size,
            self.tgt_vocab_size,
            bias=self.config.lm_head_bias,
        )
        if not checkpoint:
            init_method_configs = config._get_modules("initmethod")
            if len(init_method_configs) == 0:
                init_method = register.get("initmethod", "default")()
            else:
                assert len(init_method_configs) == 1
                init_method_config = init_method_configs[0]
                init_method = register.get(
                    "initmethod", register_module_name(init_method_config._module_name)
                )(config=init_method_config)
            if self.config.share_embedding:
                self.embedding.apply(init_method)
            else:
                self.src_embedding.apply(init_method)
                self.tgt_embedding.apply(init_method)

            self.encoder.init_weight(init_method)
            self.decoder.init_weight(init_method)
            self.lm_head.apply(init_method)
        self.pad = self.tgt_dict.pad()
        self.unk = self.tgt_dict.unk()
        self.eos = self.tgt_dict.eos()
        self.bos = self.tgt_dict.bos()
        if self.config.no_repeat_ngram_size > 0:
            self.repeat_ngram_blocker = NGramRepeatBlock(
                self.config.no_repeat_ngram_size
            )
        else:
            self.repeat_ngram_blocker = None

        assert self.config.temperature > 0, "--temperature must be greater than 0"

        token_sample_configs = config._get_modules("token_sample")
        if len(token_sample_configs) == 0:
            self.token_sample = register.get("token_sample", "beam_search")(
                self.tgt_dict
            )
        else:
            assert len(token_sample_configs) == 1
            self.token_sample = register.get(
                "token_sample",
                register_module_name(token_sample_configs[0]._module_name),
            )(self.tgt_dict, token_sample_configs[0])

        # We only need to set src_lengths in LengthConstrainedBeamSearch.
        # As a module attribute, setting it would break in multithread
        # settings when the model is shared.
        if (
            hasattr(self.token_sample, "needs_src_lengths")
            and self.token_sample.needs_src_lengths
        ):
            assert self.config.model_type == "token_enc_dec", (
                "src_lengths are only needed for token_encode_decode model"
                "but got model_type: {}".format(self.config.model_type)
            )
            self.should_set_src_lengths = True
        else:
            self.should_set_src_lengths = False

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do forward on a mini batch

        Args:
            batch: a mini batch inputs

        Returns:
            the outputs

        """
        return self.generate(inputs)

    @torch.no_grad()
    def generate(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Generate tokens.

        Args:
            inputs: one mini-batch inputs
        """
        raise NotImplementedError

    def predict_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do predict for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            the predicts outputs

        """
        return self.generate(inputs)

    def validation_step(
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """do validation for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            the validation outputs

        """
        train_step_outs = self.training_step(copy.deepcopy(inputs))
        result = self.generate(inputs)
        result[self.config.output_map.logits] = train_step_outs[
            self.config.output_map.logits
        ]
        result[self.config.output_map.decoder_target_ids] = train_step_outs[
            self.config.output_map.decoder_target_ids
        ]
        return result

    def test_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do test for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            the test outputs

        """
        train_step_outs = self.training_step(copy.deepcopy(inputs))
        result = self.generate(inputs)
        result[self.config.output_map.logits] = train_step_outs[
            self.config.output_map.logits
        ]
        result[self.config.output_map.decoder_target_ids] = train_step_outs[
            self.config.output_map.decoder_target_ids
        ]
        return result

    def training_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do training for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            the training outputs

        """
        raise NotImplementedError

    def _generate_tokens(self, bsz, beam_size, max_len, encoder_outs, src_lengths=None):
        target_device = encoder_outs[self.config.output_map.decoder_target_ids].device
        decoder_past_cache = None
        # initialize buffers
        scores = (
            torch.zeros(bsz * beam_size, max_len + 1).to(target_device).float()
        )  # +1 for eos; pad is never chosen for scoring
        tokens = (
            torch.zeros(bsz * beam_size, max_len + 2)
            .to(target_device)
            .long()
            .fill_(self.pad)
        )  # +2 for bos and pad
        tokens[:, 0] = self.bos
        attn: Optional[torch.Tensor] = None

        # A list that indicates candidates that should be ignored.
        # For example, suppose we're sampling and have already finalized 2/5
        # samples. Then cands_to_ignore would mark 2 positions as being ignored,
        # so that we only finalize the remaining 3 samples.
        cands_to_ignore = (
            torch.zeros(bsz, beam_size).to(target_device).eq(-1)
        )  # forward and backward-compatible False mask

        # list of completed sentences
        finalized = torch.jit.annotate(
            List[List[Dict[str, torch.Tensor]]],
            [torch.jit.annotate(List[Dict[str, torch.Tensor]], []) for i in range(bsz)],
        )  # contains lists of dictionaries of information about the hypothesis being finalized at each step

        # a boolean array indicating if the sentence at the index is finished or not
        finished = [False for i in range(bsz)]
        num_remaining_sent = bsz  # number of sentences remaining

        # number of candidate hypos per step
        cand_size = 2 * beam_size  # 2 x beam size in case half are EOS

        # offset arrays for converting between different indexing schemes
        bbsz_offsets = (
            (torch.arange(0, bsz) * beam_size)
            .unsqueeze(1)
            .type_as(tokens)
            .to(target_device)
        )
        cand_offsets = torch.arange(0, cand_size).type_as(tokens).to(target_device)

        reorder_state: Optional[torch.Tensor] = None
        batch_idxs: Optional[torch.Tensor] = None

        original_batch_idxs: Optional[torch.Tensor] = None
        original_batch_idxs = torch.arange(0, bsz).type_as(tokens)

        for step in range(max_len + 1):  # one extra step for EOS marker
            # reorder decoder internal states based on the prev choice of beams
            if reorder_state is not None:
                if batch_idxs is not None:
                    # update beam indices to take into account removed sentences
                    corr = batch_idxs - torch.arange(batch_idxs.numel()).type_as(
                        batch_idxs
                    )
                    reorder_state.view(-1, beam_size).add_(
                        corr.unsqueeze(-1) * beam_size
                    )
                    original_batch_idxs = original_batch_idxs[batch_idxs]
                decoder_past_cache = self.decoder.reorder_incremental_state(
                    decoder_past_cache, reorder_state
                )
                encoder_outs = self.encoder.reorder_encoder_out(
                    encoder_outs, reorder_state
                )
            encoder_outs[self.config.input_map.decoder_input_ids] = tokens[
                :, : step + 1
            ]
            if self.config.share_embedding:
                decoder_embedding = (
                    self.embedding.forward(
                        encoder_outs[self.config.input_map.decoder_input_ids][
                            :, step : step + 1
                        ]
                    )
                    * self.embedding_scale
                )
            else:
                decoder_embedding = (
                    self.tgt_embedding.forward(
                        encoder_outs[self.config.input_map.decoder_input_ids][
                            :, step : step + 1
                        ]
                    )
                    * self.embedding_scale
                )
            encoder_outs[
                self.decoder.config.input_map.decoder_input_embedding
            ] = decoder_embedding
            decoder_outs = self.decoder.forward(encoder_outs, decoder_past_cache)
            decoder_past_cache = decoder_outs[
                self.decoder.config.output_map.decoder_past_cache
            ]

            decoder_output_embedding = decoder_outs[
                self.decoder.config.output_map.decoder_output_embedding
            ]
            decoder_output_embedding = decoder_output_embedding[:, -1, :]
            lprobs = self.lm_head(decoder_output_embedding)

            avg_attn_scores = None  # TODO:get average attention score

            lprobs[lprobs != lprobs] = torch.tensor(-math.inf).to(lprobs)

            if self.pad != self.eos:
                lprobs[:, self.pad] = -math.inf  # never select pad
            if self.bos != self.eos:
                lprobs[:, self.bos] = -math.inf  # never select bos
            if self.unk != self.eos:
                lprobs[:, self.unk] -= self.config.unk_penalty  # apply unk penalty

            # handle max length constraint
            if step >= max_len:
                lprobs[:, : self.eos] = -math.inf
                lprobs[:, self.eos + 1 :] = -math.inf

            # handle prefix tokens (possibly with different lengths)
            prefix_tokens = encoder_outs.get(
                "prefix_tokens", None
            )  # TODO: prefix tokens
            if (
                prefix_tokens is not None
                and step < prefix_tokens.size(1)
                and step < max_len
            ):
                lprobs, tokens, scores = self._prefix_tokens(
                    step, lprobs, scores, tokens, prefix_tokens, beam_size
                )
            elif step < self.config.min_len:
                assert self.config.min_len < max_len
                # minimum length constraint (does not apply if using prefix_tokens)
                lprobs[:, self.eos] = -math.inf

            # Record attention scores, only support avg_attn_scores is a Tensor
            if avg_attn_scores is not None:
                if attn is None:
                    attn = torch.empty(
                        bsz * beam_size, avg_attn_scores.size(1), max_len + 2
                    ).to(scores)
                attn[:, :, step + 1].copy_(avg_attn_scores)

            scores = scores.type_as(lprobs)
            eos_bbsz_idx = torch.empty(0).to(
                tokens
            )  # indices of hypothesis ending with eos (finished sentences)
            eos_scores = torch.empty(0).to(
                scores
            )  # scores of hypothesis ending with eos (finished sentences)

            if self.should_set_src_lengths:
                self.token_sample.set_src_lengths(src_lengths)

            if self.repeat_ngram_blocker is not None:
                lprobs = self.repeat_ngram_blocker(
                    tokens[:, : step + 1], lprobs, bsz, beam_size, step
                )

            # Shape: (batch, cand_size)
            cand_scores, cand_indices, cand_beams = self.token_sample.step(
                step,
                lprobs.view(bsz, -1, self.tgt_vocab_size),
                scores.view(bsz, beam_size, -1)[:, :, :step],
                tokens[:, : step + 1],
                original_batch_idxs,
            )

            # cand_bbsz_idx contains beam indices for the top candidate
            # hypotheses, with a range of values: [0, bsz*beam_size),
            # and dimensions: [bsz, cand_size]
            cand_bbsz_idx = cand_beams.add(bbsz_offsets)

            # finalize hypotheses that end in eos
            # Shape of eos_mask: (batch size, beam size)
            eos_mask = cand_indices.eq(self.eos) & cand_scores.ne(-math.inf)
            eos_mask[:, :beam_size][cands_to_ignore] = torch.tensor(0).to(eos_mask)

            # only consider eos when it's among the top beam_size indices
            # Now we know what beam item(s) to finish
            # Shape: 1d list of absolute-numbered
            eos_bbsz_idx = torch.masked_select(
                cand_bbsz_idx[:, :beam_size], mask=eos_mask[:, :beam_size]
            )

            finalized_sents: List[int] = []
            if eos_bbsz_idx.numel() > 0:
                eos_scores = torch.masked_select(
                    cand_scores[:, :beam_size], mask=eos_mask[:, :beam_size]
                )

                finalized_sents = self.finalize_hypos(
                    step,
                    eos_bbsz_idx,
                    eos_scores,
                    tokens,
                    scores,
                    finalized,
                    finished,
                    beam_size,
                    attn,
                    src_lengths,
                    max_len,
                )
                num_remaining_sent -= len(finalized_sents)

            assert num_remaining_sent >= 0
            if num_remaining_sent == 0:
                break

            # Remove finalized sentences (ones for which {beam_size}
            # finished hypotheses have been generated) from the batch.
            if len(finalized_sents) > 0:
                new_bsz = bsz - len(finalized_sents)

                # construct batch_idxs which holds indices of batches to keep for the next pass
                batch_mask = torch.ones(
                    bsz, dtype=torch.bool, device=cand_indices.device
                )
                batch_mask[finalized_sents] = False
                # TODO replace `nonzero(as_tuple=False)` after TorchScript supports it
                batch_idxs = torch.arange(
                    bsz, device=cand_indices.device
                ).masked_select(batch_mask)

                # Choose the subset of the hypothesized constraints that will continue
                self.token_sample.prune_sentences(batch_idxs)

                eos_mask = eos_mask[batch_idxs]
                cand_beams = cand_beams[batch_idxs]
                bbsz_offsets.resize_(new_bsz, 1)
                cand_bbsz_idx = cand_beams.add(bbsz_offsets)
                cand_scores = cand_scores[batch_idxs]
                cand_indices = cand_indices[batch_idxs]

                if prefix_tokens is not None:
                    prefix_tokens = prefix_tokens[batch_idxs]
                if src_lengths is not None:
                    src_lengths = src_lengths[batch_idxs]
                cands_to_ignore = cands_to_ignore[batch_idxs]

                scores = scores.view(bsz, -1)[batch_idxs].view(new_bsz * beam_size, -1)
                tokens = tokens.view(bsz, -1)[batch_idxs].view(new_bsz * beam_size, -1)
                if attn is not None:
                    attn = attn.view(bsz, -1)[batch_idxs].view(
                        new_bsz * beam_size, attn.size(1), -1
                    )
                bsz = new_bsz
            else:
                batch_idxs = None

            # Set active_mask so that values > cand_size indicate eos hypos
            # and values < cand_size indicate candidate active hypos.
            # After, the min values per row are the top candidate active hypos

            # Rewrite the operator since the element wise or is not supported in torchscript.

            eos_mask[:, :beam_size] = ~((~cands_to_ignore) & (~eos_mask[:, :beam_size]))
            active_mask = torch.add(
                eos_mask.type_as(cand_offsets) * cand_size,
                cand_offsets[: eos_mask.size(1)],
            )

            # get the top beam_size active hypotheses, which are just
            # the hypos with the smallest values in active_mask.
            # {active_hypos} indicates which {beam_size} hypotheses
            # from the list of {2 * beam_size} candidates were
            # selected. Shapes: (batch size, beam size)
            new_cands_to_ignore, active_hypos = torch.topk(
                active_mask, k=beam_size, dim=1, largest=False
            )

            # update cands_to_ignore to ignore any finalized hypos.
            cands_to_ignore = new_cands_to_ignore.ge(cand_size)[:, :beam_size]
            # Make sure there is at least one active item for each sentence in the batch.
            assert (~cands_to_ignore).any(dim=1).all()

            # update cands_to_ignore to ignore any finalized hypos

            # {active_bbsz_idx} denotes which beam number is continued for each new hypothesis (a beam
            # can be selected more than once).
            active_bbsz_idx = torch.gather(cand_bbsz_idx, dim=1, index=active_hypos)
            active_scores = torch.gather(cand_scores, dim=1, index=active_hypos)

            active_bbsz_idx = active_bbsz_idx.view(-1)
            active_scores = active_scores.view(-1)

            # copy tokens and scores for active hypotheses

            # Set the tokens for each beam (can select the same row more than once)
            tokens[:, : step + 1] = torch.index_select(
                tokens[:, : step + 1], dim=0, index=active_bbsz_idx
            )
            # Select the next token for each of them
            tokens.view(bsz, beam_size, -1)[:, :, step + 1] = torch.gather(
                cand_indices, dim=1, index=active_hypos
            )
            if step > 0:
                scores[:, :step] = torch.index_select(
                    scores[:, :step], dim=0, index=active_bbsz_idx
                )
            scores.view(bsz, beam_size, -1)[:, :, step] = torch.gather(
                cand_scores, dim=1, index=active_hypos
            )

            # Update constraints based on which candidates were selected for the next beam
            self.token_sample.update_constraints(active_hypos)

            # copy attention for active hypotheses
            if attn is not None:
                attn[:, :, : step + 2] = torch.index_select(
                    attn[:, :, : step + 2], dim=0, index=active_bbsz_idx
                )

            # reorder incremental state in decoder
            reorder_state = active_bbsz_idx

        # sort by score descending
        for sent in range(len(finalized)):
            scores = torch.tensor(
                [float(elem["score"].item()) for elem in finalized[sent]]
            )
            _, sorted_scores_indices = torch.sort(scores, descending=True)
            finalized[sent] = [finalized[sent][ssi] for ssi in sorted_scores_indices]
            finalized[sent] = torch.jit.annotate(
                List[Dict[str, torch.Tensor]], finalized[sent]
            )
        return finalized

    def _prefix_tokens(
        self, step: int, lprobs, scores, tokens, prefix_tokens, beam_size: int
    ):
        """Handle prefix tokens"""
        prefix_toks = prefix_tokens[:, step].unsqueeze(-1).repeat(1, beam_size).view(-1)
        prefix_lprobs = lprobs.gather(-1, prefix_toks.unsqueeze(-1))
        prefix_mask = prefix_toks.ne(self.pad)
        lprobs[prefix_mask] = torch.tensor(-math.inf).to(lprobs)
        lprobs[prefix_mask] = lprobs[prefix_mask].scatter(
            -1, prefix_toks[prefix_mask].unsqueeze(-1), prefix_lprobs[prefix_mask]
        )
        # if prefix includes eos, then we should make sure tokens and
        # scores are the same across all beams
        eos_mask = prefix_toks.eq(self.eos)
        if eos_mask.any():
            # validate that the first beam matches the prefix
            first_beam = tokens[eos_mask].view(-1, beam_size, tokens.size(-1))[
                :, 0, 1 : step + 1
            ]
            eos_mask_batch_dim = eos_mask.view(-1, beam_size)[:, 0]
            target_prefix = prefix_tokens[eos_mask_batch_dim][:, :step]
            assert (first_beam == target_prefix).all()

            # copy tokens, scores and lprobs from the first beam to all beams
            tokens = self.replicate_first_beam(tokens, eos_mask_batch_dim, beam_size)
            scores = self.replicate_first_beam(scores, eos_mask_batch_dim, beam_size)
            lprobs = self.replicate_first_beam(lprobs, eos_mask_batch_dim, beam_size)
        return lprobs, tokens, scores

    def replicate_first_beam(self, tensor, mask, beam_size: int):
        tensor = tensor.view(-1, beam_size, tensor.size(-1))
        tensor[mask] = tensor[mask][:, :1, :]
        return tensor.view(-1, tensor.size(-1))

    def finalize_hypos(
        self,
        step: int,
        bbsz_idx,
        eos_scores,
        tokens,
        scores,
        finalized: List[List[Dict[str, torch.Tensor]]],
        finished: List[bool],
        beam_size: int,
        attn: Optional[torch.Tensor],
        src_lengths,
        max_len: int,
    ):
        """Finalize hypothesis, store finalized information in `finalized`, and change `finished` accordingly.
        A sentence is finalized when {beam_size} finished items have been collected for it.

        Returns number of sentences (not beam items) being finalized.
        These will be removed from the batch and not processed further.
        Args:
            bbsz_idx (Tensor):
        """
        assert bbsz_idx.numel() == eos_scores.numel()

        # clone relevant token and attention tensors.
        # tokens is (batch * beam, max_len). So the index_select
        # gets the newly EOS rows, then selects cols 1..{step + 2}
        tokens_clone = tokens.index_select(0, bbsz_idx)[
            :, 1 : step + 2
        ]  # skip the first index, which is BOS

        tokens_clone[:, step] = self.eos
        attn_clone = (
            attn.index_select(0, bbsz_idx)[:, :, 1 : step + 2]
            if attn is not None
            else None
        )

        # compute scores per token position
        pos_scores = scores.index_select(0, bbsz_idx)[:, : step + 1]
        pos_scores[:, step] = eos_scores
        # convert from cumulative to per-position scores
        pos_scores[:, 1:] = pos_scores[:, 1:] - pos_scores[:, :-1]

        # normalize sentence-level scores
        if self.config.normalize_scores:
            eos_scores /= (step + 1) ** self.config.len_penalty

        # cum_unfin records which sentences in the batch are finished.
        # It helps match indexing between (a) the original sentences
        # in the batch and (b) the current, possibly-reduced set of
        # sentences.
        cum_unfin: List[int] = []
        prev = 0
        # [1, 0, 1, 0, 0, 1]
        # [1, 2, 2]
        for f in finished:
            if f:
                prev += 1
            else:
                cum_unfin.append(prev)
        cum_fin_tensor = torch.tensor(cum_unfin, dtype=torch.int).to(bbsz_idx)

        unfin_idx = torch.div(bbsz_idx, beam_size, rounding_mode="trunc")
        sent = unfin_idx + torch.index_select(cum_fin_tensor, 0, unfin_idx)

        # Create a set of "{sent}{unfin_idx}", where
        # "unfin_idx" is the index in the current (possibly reduced)
        # list of sentences, and "sent" is the index in the original,
        # unreduced batch
        # For every finished beam item
        # sentence index in the current (possibly reduced) batch
        seen = (sent << 32) + unfin_idx
        unique_seen: List[int] = torch.unique(seen).tolist()

        # NOTE: only for token encode and token decode model should we consider the match_source_len
        if (
            self.config.model_type == "token_encode_decode"
            and self.config.match_source_len
        ):
            condition = step > torch.index_select(src_lengths, 0, unfin_idx)
            eos_scores = torch.where(condition, torch.tensor(-math.inf), eos_scores)
        sent_list: List[int] = sent.tolist()
        for i in range(bbsz_idx.size()[0]):
            # An input sentence (among those in a batch) is finished when
            # beam_size hypotheses have been collected for it
            if len(finalized[sent_list[i]]) < beam_size:
                if attn_clone is not None:
                    # remove padding tokens from attn scores
                    hypo_attn = attn_clone[i]
                else:
                    hypo_attn = torch.empty(0)

                finalized[sent_list[i]].append(
                    {
                        "tokens": tokens_clone[i],
                        "score": eos_scores[i],
                        "attention": hypo_attn,  # src_len x tgt_len
                        "alignment": torch.empty(0),
                        "positional_scores": pos_scores[i],
                    }
                )

        newly_finished: List[int] = []
        for unique_s in unique_seen:
            # check termination conditions for this sentence
            unique_sent: int = unique_s >> 32
            unique_unfin_idx: int = unique_s - (unique_sent << 32)

            if not finished[unique_sent] and self.is_finished(
                step, unique_unfin_idx, max_len, len(finalized[unique_sent]), beam_size
            ):
                finished[unique_sent] = True
                newly_finished.append(unique_unfin_idx)

        return newly_finished

    def is_finished(
        self,
        step: int,
        unfin_idx: int,
        max_len: int,
        finalized_sent_len: int,
        beam_size: int,
    ):
        """
        Check whether decoding for a sentence is finished, which
        occurs when the list of finalized sentences has reached the
        beam size, or when we reach the maximum length.
        """
        assert finalized_sent_len <= beam_size
        if finalized_sent_len == beam_size or step == max_len:
            return True
        return False



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Any, Callable, Dict, List, Union

import hjson
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    Parser,
    StrField,
    SubModule,
    cregister,
    init_config,
)
from intc.utils import fix_trace

import dlk.data.dataset
import dlk.nn
from dlk.utils.io import open
from dlk.utils.register import register, register_module_name



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict

from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from lightning.pytorch.callbacks import LearningRateMonitor

from dlk.utils.register import register


@cregister("callback", "lr_monitor")
class LearningRateMonitorCallbackConfig(Base):
    """callback for monitor the learning rate of the optimizer"""

    logging_interval = StrField(
        value=None,
        options=["step", "epoch", None],
        help="""set to ``'epoch'`` or ``'step'`` to log ``lr`` of all optimizers at the same interval, set to ``None`` to log at individual interval according to the ``interval`` key of each scheduler. Defaults to ``None``.""",
    )
    log_momentum = BoolField(
        value=False,
        help="option to also log the momentum values of the optimizer, if the optimizer has the ``momentum`` or ``betas`` attribute. Defaults to ``False``.",
    )


@register("callback", "lr_monitor")
class LearningRateMonitorCallback(object):
    """Monitor the learning rate"""

    def __init__(self, config: LearningRateMonitorCallbackConfig):
        super().__init__()
        self.config = config._to_dict(only_para=True)

    def __call__(self, rt_config: Dict) -> LearningRateMonitor:
        """return LearningRateMonitor object

        Args:
            rt_config: runtime config, include save_dir, and the checkpoint path name

        Returns:
            LearningRateMonitor object

        """
        return LearningRateMonitor(**self.config)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import pickle as pkl
from typing import Dict, List, Set

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import SimpleModule
from dlk.nn.layer.embedding.random import RandomEmbedding, RandomEmbeddingConfig
from dlk.utils.io import open
from dlk.utils.register import register

logger = logging.getLogger(__name__)


@cregister("embedding", "static")
class StaticEmbeddingConfig(RandomEmbeddingConfig):
    """
    the static embedding module config
    """

    embedding_file = StrField(
        value=MISSING,
        help="the embedding file, must be saved as numpy array by pickle",
    )
    freeze = BoolField(value=False, help="whether to freeze the embedding")


@register("embedding", "static")
class StaticEmbedding(RandomEmbedding):
    """from 'input_ids' generate static 'embedding' like glove, word2vec"""

    def __init__(self, config: StaticEmbeddingConfig):
        super().__init__(config)
        self.config = config
        embedding_file = self.config.embedding_file
        with open(embedding_file, "rb") as f:
            embedding_file = pkl.load(f)
        self.dropout = nn.Dropout(float(self.config.dropout))
        self.embedding = nn.Embedding.from_pretrained(
            torch.tensor(embedding_file, dtype=torch.float),
            freeze=self.config.freeze,
            padding_idx=self.config.padding_idx,
        )
        assert (
            self.embedding.weight.shape[-1] == self.config.embedding_dim
        ), f"{self.embedding.weight.shape[-1]} != {self.config.embedding_dim}"

    def init_weight(self, method):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        logger.info(f"The static embedding is loaded the pretrained embedding.")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.


import numpy as np
import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import Module


@cregister("module", "biaffine")
class BiAffineConfig:
    """the biaffine module config"""

    input_size = IntField(
        value=MISSING,
        minimum=0,
        help="the input size of the biaffine module",
    )
    hidden_size = IntField(
        value=0,
        minimum=0,
        help="the hidden size of the biaffine module, if set to 0, will set the hidden size to the input size",
    )
    output_size = IntField(
        value=MISSING,
        minimum=0,
        help="the output size of the biaffine module",
    )
    dropout = FloatField(
        value=0.0,
        minimum=0.0,
        maximum=1.0,
        help="the dropout rate of the biaffine module",
    )
    multi_matrix = IntField(
        value=1, minimum=1, help="the number of the matrix of return result"
    )
    max_seq_len = IntField(
        value=1024,
        minimum=0,
        help="the max sequence length of the biaffine module",
    )
    bias = BoolField(value=True, help="whether to use the bias in the biaffine module")
    active = StrField(
        value="leaky_relu",
        options=["none", "leaky_relu", "relu", "gelu", "glu", "selu", "celu", "elu"],
        help="the activation function",
    )


active_map = {
    "leaky_relu": nn.LeakyReLU,
    "relu": nn.ReLU,
    "gelu": nn.GELU,
    "glu": nn.GLU,
    "selu": nn.SELU,
    "celu": nn.CELU,
    "elu": nn.ELU,
}


@register("module", "biaffine")
class BiAffine(Module):
    """wrap for nn.BiAffine"""

    def __init__(self, config: BiAffineConfig):
        super(BiAffine, self).__init__()
        self.config = config
        if self.config.hidden_size == 0:
            self.config.hidden_size = self.config.input_size
        self.linear_a = nn.Linear(self.config.input_size, self.config.hidden_size)
        self.linear_b = nn.Linear(self.config.input_size, self.config.hidden_size)
        self.dropout = nn.Dropout(p=self.config.dropout)
        if config.active != "none":
            self.active = active_map[config.active]()
        else:
            self.active = lambda x: x
        if self.config.bias:
            self.biaffine = nn.Parameter(
                torch.randn(
                    self.config.hidden_size + 1,
                    self.config.output_size * self.config.multi_matrix,
                    self.config.hidden_size + 1,
                )
            )
        else:
            self.biaffine = nn.Parameter(
                torch.randn(
                    self.config.hidden_size,
                    self.config.output_size * self.config.multi_matrix,
                    self.config.hidden_size,
                )
            )

    def init_weight(self, method):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        torch.nn.init.xavier_uniform_(self.biaffine)
        self.linear_a.apply(method)
        self.linear_b.apply(method)

    def forward(self, embedding: torch.Tensor) -> torch.Tensor:
        """do forward on a mini batch

        Args:
            embedding: a mini batch embedding, shape==(batch_size, input_a_len, input_size)

        Returns:
            input_a x biaffine x input_b, shape==(batch_size, input_a_len, input_b_len, output_size)

        """

        input_a = self.dropout(self.active(embedding))
        input_b = self.dropout(self.active(embedding))
        input_a = self.linear_a(input_a)
        input_b = self.linear_b(input_b)

        if self.config.bias:
            output = torch.einsum(
                "bmi,ioj,bnj->bmno",
                torch.cat((input_a, torch.ones_like(input_a[..., :1])), dim=-1),
                self.biaffine,
                torch.cat((input_b, torch.ones_like(input_b[..., :1])), dim=-1),
            )
        else:
            output = torch.einsum(
                "bmi,ioj,bnj->bmno",
                input_a,
                self.biaffine,
                input_b,
            )
        if self.config.multi_matrix > 1:
            bs, seq_len, _, output_size = output.shape
            output = output.reshape(
                bs, seq_len, seq_len, self.config.multi_matrix, self.config.output_size
            )
            output = output.permute(
                0, 3, 1, 2, 4
            )  # bs, group, seq_len, seq_len, output_size)
        return output



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.


import json
import logging
import os
from typing import Dict, List, Optional

import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from tokenizers import Tokenizer
from torchmetrics.functional.text import bleu_score

from dlk.data.postprocessor import BasePostProcessor, BasePostProcessorConfig
from dlk.utils.register import register, register_module_name

logger = logging.getLogger(__name__)


@cregister("postprocessor", "token_generate")
class TokenGeneratePostProcessorConfig(BasePostProcessorConfig):
    """token generate postprocessor config"""

    class InputMap:
        logits = StrField(value="logits", help="the output logits")
        decoder_target_ids = StrField(value="decoder_target_ids")
        generated = StrField(value="generated")
        index = StrField(value="_index", help="the index of the sample")

    skip_special_tokens = BoolField(
        value=True, help="When decode the tokens, skip the special tokens"
    )
    tokenizer = StrField(value=MISSING, help="the tokenizer config file path")
    return_all_generations = BoolField(
        value=False, help="return all generations, or only the top one"
    )

    class OriginInputMap:
        uuid = StrField(value="uuid", help="the uuid or the id of the sample")
        input = StrField(value="input", help="the prompt for generation")
        target = StrField(value="target", help="the target generation")

    origin_input_map = NestField(
        value=OriginInputMap,
        help="""
        the origin input map of the processor,
        the key is the name of the processor needed key,
        the value is the provided data provided key
        """,
    )

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )


@register("postprocessor", "token_generate")
class TokenGeneratePostProcessor(BasePostProcessor):
    """token generate postprocess"""

    def __init__(self, config: TokenGeneratePostProcessorConfig):
        super(TokenGeneratePostProcessor, self).__init__(config)
        self.config = config

        self.tokenizer = Tokenizer.from_file(self.config.tokenizer)

    def _get_origin_data(self, one_origin: pd.Series) -> Dict:
        """

        Args:
            one_origin: the original data

        Returns:
            the gather origin data

        """
        origin = {}
        origin["input"] = one_origin[self.config.origin_input_map.input]
        origin["uuid"] = one_origin[self.config.origin_input_map.uuid]
        origin["target"] = one_origin[self.config.origin_input_map.target]
        return origin

    def do_predict(
        self,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> List:
        """Process the model predict to human readable format

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            all predicts
        """
        results = []
        for outputs in list_batch_outputs:
            indexes = list(outputs[self.config.input_map.index])

            batch_generated = outputs[self.config.input_map.generated]
            for i, (index, generated) in enumerate(zip(indexes, batch_generated)):
                one_origin = origin_data.iloc[int(index)]
                one_ins = self._get_origin_data(one_origin)

                generate_result = []
                for i, (one_generate) in enumerate(generated):
                    generate_sent = self.tokenizer.decode(
                        list(one_generate["tokens"]),
                        skip_special_tokens=self.config.skip_special_tokens,
                    )
                    generate_result.append(
                        {
                            "generate": generate_sent,
                            "score": float(one_generate["score"]),
                        }
                    )
                    if self.config.return_all_generations == False:
                        break
                one_ins["generated"] = generate_result
                one_ins["predict_extend_return"] = self.gather_predict_extend_data(
                    outputs, i, self.config.predict_extend_return
                )
                results.append(one_ins)
        return results

    def do_calc_metrics(
        self,
        predicts: List,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> Dict:
        """calc the scores use the predicts or list_batch_outputs

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            the named scores

        """
        generates = []
        targets = []
        for one_ins in predicts:
            generates.append(one_ins["generated"][0]["generate"])
            targets.append(one_ins["target"])
        bleu_score_value = float(bleu_score(generates, targets))

        real_name = self.loss_name_map(stage)
        return {f"{real_name}_bleu": bleu_score_value}



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import copy
import os
from typing import Any, Dict

import pandas as pd
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import Dataset

from dlk.utils.register import register


@cregister("dataset", "default")
class DefaultDatasetConfig(Base):
    """the default dataset"""

    key_type_pairs = DictField(value={}, help="the pair of key and type")
    repeat_for_valid = BoolField(
        value=True, help="whether to repeat the data for valid"
    )


@register("dataset", "default")
class DefaultDataset(Dataset):
    """General Dataset"""

    def __init__(
        self,
        config: DefaultDatasetConfig,
        data: pd.DataFrame,
        rt_config: Dict,
        key_type_pairs: Dict = None,
    ):
        """
        if repeat_valid >1 we will repeat each item in this dataset
        """
        self.config = config

        self.repeat_valid = 1
        if self.config.repeat_for_valid and rt_config.get("world_size", 1) > 1:
            self.repeat_valid = rt_config.get("world_size", 1)
        self.data = data
        self.type_map = {
            "float": torch.float,
            "int": torch.int,
            "bool": torch.bool,
            "long": torch.long,
        }
        if key_type_pairs is not None:
            self.key_type_pairs = key_type_pairs
        else:
            self.key_type_pairs = self.real_key_type_pairs(config.key_type_pairs, data)

    @staticmethod
    def real_key_type_pairs(key_type_pairs: Dict, data: pd.DataFrame):
        """return the keys = key_type_pairs.keys() ∩ data.columns

        Args:
            key_type_pairs: data in columns should map to tensor type
            data: the pd.DataFrame

        Returns:
            real_key_type_pairs where keys = key_type_pairs.keys() ∩ data.columns

        """
        copy_key_type_pairs = copy.deepcopy(key_type_pairs)
        has_key = set(data.columns)
        remove = set()
        for key in copy_key_type_pairs:
            if key not in has_key:
                remove.add(key)
        for key in remove:
            copy_key_type_pairs.pop(key)
        return copy_key_type_pairs

    def __len__(self):
        """return the dataset size"""
        return len(self.data) * self.repeat_valid

    def __getitem__(self, idx: int):
        """return one instance by index

        Args:
            idx: the index of data

        Returns:
            the data[idx] and convert to tensor the result will add 'idx' to '_index'

        """
        idx = idx // self.repeat_valid
        one_ins = {}
        for key, key_type in self.key_type_pairs.items():
            one_ins[key] = torch.tensor(
                self.data.iloc[idx][key], dtype=self.type_map[key_type]
            )
        one_ins["_index"] = torch.tensor(idx, dtype=torch.long)
        return one_ins



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Dict

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from packaging import version

from dlk.utils.register import register

from . import BaseLoss, BaseLossConfig

logger = logging.getLogger(__name__)


@cregister("loss", "cross_entropy")
class CrossEntropyLossConfig(BaseLossConfig):
    """the cross_entropy loss"""

    weight = ListField(
        value=None,
        additions=[None],
        help="the list of weights of every class",
    )
    label_smoothing = FloatField(
        value=0.0, minimum=0.0, maximum=1.0, help="the label smoothing"
    )
    ignore_index = IntField(value=-100, help="the ignore index")


@register("loss", "cross_entropy")
class CrossEntropyLoss(BaseLoss):
    """for multi class classification"""

    def __init__(self, config: CrossEntropyLossConfig):
        super(CrossEntropyLoss, self).__init__(config)
        self.config: CrossEntropyLossConfig
        if self.config.weight:
            weight = torch.tensor(self.config.weight, dtype=torch.float)
        else:
            weight = None
        if version.parse(torch.__version__) >= version.parse("1.10"):
            self.cross_entropy = nn.CrossEntropyLoss(
                weight=weight,
                ignore_index=self.config.ignore_index,
                reduction=self.config.reduction,
                label_smoothing=self.config.label_smoothing,
            )
        else:
            if self.config.label_smoothing:
                logger.info("Torch version is <1.10, so ignore label_smoothing")
            self.cross_entropy = nn.CrossEntropyLoss(
                weight=weight,
                ignore_index=self.config.ignore_index,
                reduction=self.config.reduction,
            )

    def _calc(self, result, inputs, rt_config, scale):
        """calc the loss the predict is from result, the ground truth is from inputs

        Args:
            result: the model predict dict
            inputs: the all inputs for model
            rt_config: provide the current training status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            scale: the scale rate for the loss
        Returns:
            loss

        """
        pred = result[self.pred_name]
        target = inputs[self.truth_name]
        pred = pred.reshape(-1, pred.shape[-1])
        target = target.reshape(-1)
        loss = self.cross_entropy(pred, target) * scale
        return loss, {self.config.log_map.loss: loss}



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import dlk.nn.layer
import dlk.nn.model
import dlk.nn.module
import dlk.nn.token_sample



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import copy
import logging
from functools import lru_cache
from typing import Callable, Dict, List, Union

import lightning as pl
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    asdict,
    cregister,
)

from dlk.utils.register import register, register_module_name

logger = logging.getLogger(__name__)


def check_accept_submodule(submodule_config: Dict):
    """check the submodule config is valid
    Args:
        submodule_config:
            the submodule config, it should be a dict, the key is the submodule name
    Returns:
        None
    Raises:
        KeyError: if the submodule name is not accept
    """
    accept_submodule = [
        "model",
        "optimizer",
        "scheduler",
        "loss",
        "postprocessor",
        "adv_method",
    ]
    for key in submodule_config:
        key = key.split("@")[0].split("#")[0].split("-")[0]
        if key not in accept_submodule:
            logger.error(
                f"The submodule name {key} is not accept for default IModel, please use one of {accept_submodule}."
            )
            return False
    return True


@cregister("imodel", "default")
class DefaultIModelConfig(Base):
    """
    The Most General IModel, include the model, optimizer, scheduler, loss, postprocessor
    """

    gather_metrics = BoolField(
        value=True,
        help="when using the distributed training, whether gather all the result to rank 0 to calculate the metrics, if `False` we will only calculate the metrics which on rank 0.",
    )
    validate_names = ListField(
        value=["val"],
        help="the name(s) of the validation datasets(if there are more than one validation dataset, you should provide the names for each one), used in the training process",
    )
    test_names = ListField(
        value=["test"],
        help="the name(s) of the test datasets(if there are more than one test dataset, you should provide the names for each one), used in the training process",
    )

    submodule = SubModule(
        value={},
        suggestions=[
            "model",
            "optimizer",
            "scheduler",
            "loss",
            "postprocessor",
            "adv_method",
        ],
        validator=check_accept_submodule,
    )


@register("imodel", "default")
class DefaultIModel(pl.LightningModule):
    """ """

    def __init__(
        self, config: DefaultIModelConfig, checkpoint=False, rt_config: Dict = {}
    ):
        """init all modules except scheduler which requires the information from datamodule(training steps and every epoch steps)"""
        super().__init__()
        self.config = config
        model_configs = config._get_modules("model")
        assert (
            len(model_configs) == 1
        ), "The model submodule should only have one config"
        self.model = register.get(
            "model", register_module_name(model_configs[0]._module_name)
        )(model_configs[0], checkpoint)

        self.train_rt_config = rt_config
        loss_configs = config._get_modules("loss")
        assert len(loss_configs) == 1, "The loss submodule should only have one config"
        self.calc_loss = register.get(
            "loss", register_module_name(loss_configs[0]._module_name)
        )(loss_configs[0])

        postprocessor_configs = config._get_modules("postprocessor")
        assert (
            len(postprocessor_configs) == 1
        ), "The postprocessor submodule should only have one config"
        self.postprocessor = register.get(
            "postprocessor", register_module_name(postprocessor_configs[0]._module_name)
        )(postprocessor_configs[0])

        adv_method_configs = config._get_modules("adv_method")

        if len(adv_method_configs) == 1 and adv_method_configs[0].use_adv:
            self.adv_method = register.get(
                "adv_method", register_module_name(adv_method_configs[0]._module_name)
            )(self.model, adv_method_configs[0])
            self.automatic_optimization = False
        else:
            self.adv_method = None
        self.validation_outputs = {}
        self.test_outputs = {}
        self._reset_output("valid")
        self._reset_output("test")

        self._origin_data = {}
        self.gather_data: Dict = asdict(postprocessor_configs[0].input_map)
        self.gather_data.update(postprocessor_configs[0].predict_extend_return)

    def _reset_output(self, stage: str):
        """reset the validation and test output to empty
        Returns:
            None
        """
        if stage == "valid":
            for i, name in enumerate(self.config.validate_names):
                self.validation_outputs[i] = {"name": name, "outputs": []}
        elif stage == "test":
            for i, name in enumerate(self.config.test_names):
                self.test_outputs[i] = {"name": name, "outputs": []}
        else:
            raise ValueError(f"stage should be valid or test, but got {stage}")

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do forward on a mini batch

        Args:
            batch: a mini batch inputs

        Returns:
            the outputs

        """
        return self.model(inputs)

    def training_step(self, batch: Dict[str, torch.Tensor], batch_idx: int):
        """do training_step on a mini batch

        Args:
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch

        Returns:
            the outputs

        """
        if self.adv_method:
            loss, loss_log = self.adv_method.training_step(self, batch, batch_idx)
        else:
            result = self.model.training_step(batch)
            loss, loss_log = self.calc_loss(
                result,
                batch,
                rt_config={
                    "current_step": self.global_step,
                    "current_epoch": self.current_epoch,
                    "total_steps": self.num_training_steps,
                    "total_epochs": self.num_training_epochs,
                },
            )
        log_info = {}
        for key in loss_log:
            log_info[f"train_{key}"] = loss_log[key].unsqueeze(0).detach().cpu()
        self.log_dict(log_info, prog_bar=True, logger=True)
        return loss

    def validation_step(
        self, batch: Dict[str, torch.Tensor], batch_idx: int, dataloader_idx: int = 0
    ) -> Dict[str, torch.Tensor]:
        """do validation on a mini batch

        The outputs only gather the keys in self.gather_data.keys for postprocess
        Args:
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch
            dataloader_idx: the index of the multi dataloaders

        Returns:
            the outputs

        """
        result = self.model.validation_step(batch)
        loss, loss_log = self.calc_loss(
            result,
            batch,
            rt_config={  # align with training step
                "current_step": self.global_step,
                "current_epoch": self.current_epoch,
                "total_steps": self.num_training_steps,
                "total_epochs": self.num_training_epochs,
            },
        )
        log_info = {}
        for key in loss_log:
            log_info[f"{key}"] = loss_log[key].unsqueeze(0).detach().cpu()
        gather_column = list(self.gather_data.keys())
        return_result = log_info  # this loss will be used in postprocess
        for column in gather_column:
            column = self.gather_data[column]
            if column in result:
                return_result[column] = result[column]
        return_result["_index"] = batch["_index"]

        self.validation_outputs[dataloader_idx]["outputs"].append(return_result)
        return return_result

    def on_validation_epoch_end(self) -> None:
        """Gather the outputs of all node and do postprocess on it.

        The outputs only gather the keys in self.gather_data.keys for postprocess

        Returns:
            None

        """
        if self.config.gather_metrics:
            for index, validation_output in self.validation_outputs.items():
                if self.trainer.world_size > 1:
                    outputs_list = [[] for _ in range(self.trainer.world_size)]
                    torch.distributed.all_gather_object(
                        outputs_list, validation_output["outputs"]
                    )
                    outputs = []
                    for output in outputs_list:
                        outputs.extend(output)
                else:
                    outputs = validation_output["outputs"]
                self.validation_outputs[index]["outputs"] = outputs

        for index, validation_output in self.validation_outputs.items():
            log_dict = self.postprocessor(
                stage="valid",
                list_batch_outputs=validation_output["outputs"],
                origin_data=self._origin_data["valid"],
                rt_config={
                    "current_step": self.global_step,
                    "current_epoch": self.current_epoch,
                    "total_steps": self.num_training_steps,
                    "total_epochs": self.num_training_epochs,
                    "log_name": validation_output["name"],
                    "log_dir": self.train_rt_config["log_dir"],
                    "name": self.train_rt_config["name"],
                },
            )
            self.log_dict(
                log_dict,
                prog_bar=True,
                rank_zero_only=True,
            )
            if self.trainer.loggers and self.train_rt_config["hp_metrics"] in log_dict:
                hp_met = self.train_rt_config["hp_metrics"]
                self.trainer.loggers[0].log_hyperparams(
                    self.train_rt_config["hyper_config"],
                    metrics={hp_met: log_dict[hp_met]},
                )
        self._reset_output("valid")
        return None

    def test_step(
        self, batch: Dict[str, torch.Tensor], batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        """do test on a mini batch

        The outputs only gather the keys in self.gather_data.keys for postprocess
        Args:
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch
            dataloader_idx: the index of the multi dataloaders

        Returns:
            None

        """
        result = self.model.test_step(batch)
        loss, loss_log = self.calc_loss(
            result,
            batch,
            rt_config={  # align with training step
                "current_step": self.global_step,
                "current_epoch": self.current_epoch,
                "total_steps": self.num_training_steps,
                "total_epochs": self.num_training_epochs,
            },
        )
        log_info = {}
        for key in loss_log:
            log_info[f"{key}"] = loss_log[key].unsqueeze(0)
        gather_column = list(self.gather_data.keys())
        return_result = log_info  # this loss will use in postprocess
        for column in gather_column:
            if column in result:
                return_result[column] = result[column]
        return_result["_index"] = batch["_index"]
        self.test_outputs[dataloader_idx]["outputs"].append(return_result)
        return None

    def on_test_epoch_end(self) -> None:
        """Gather the outputs of all node and do postprocess on it.

        Returns:
            None
        """
        if self.config.gather_metrics:
            for index, test_output in self.test_outputs.items():
                if self.trainer.world_size > 1:
                    outputs_list = [[] for _ in range(self.trainer.world_size)]
                    torch.distributed.all_gather_object(
                        outputs_list, test_output["outputs"]
                    )
                    outputs = []
                    for output in outputs_list:
                        outputs.extend(output)
                else:
                    outputs = test_output["outputs"]
                self.test_outputs[index]["outputs"] = outputs

        for index, test_output in self.test_outputs.items():
            log_dict = self.postprocessor(
                stage="test",
                list_batch_outputs=test_output["outputs"],
                origin_data=self._origin_data["test"],
                rt_config={
                    "current_step": self.global_step,
                    "current_epoch": self.current_epoch,
                    "total_steps": self.num_training_steps,
                    "total_epochs": self.num_training_epochs,
                    "log_name": test_output["name"],
                    "log_dir": self.train_rt_config["log_dir"],
                    "name": self.train_rt_config["name"],
                },
            )
            self.log_dict(
                log_dict,
                prog_bar=True,
                rank_zero_only=True,
            )
            if self.trainer.loggers and self.train_rt_config["hp_metrics"] in log_dict:
                hp_met = self.train_rt_config["hp_metrics"]
                self.trainer.loggers[0].log_hyperparams(
                    self.train_rt_config["hyper_config"],
                    metrics={hp_met: log_dict[hp_met]},
                )
        self._reset_output("test")
        return None

    def predict_step(self, batch: Dict, batch_idx: int) -> Dict:
        """do predict on a mini batch

        Args:
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch

        Returns:
            the outputs

        """
        result = self.model.predict_step(batch)
        gather_column = list(self.gather_data.keys())
        return_result = {}
        for column in gather_column:
            column = self.gather_data[column]
            if column in result:
                return_result[column] = result[column]
        return_result["_index"] = batch["_index"]
        return return_result

    @property
    @lru_cache(maxsize=5)  # the size should always == 1
    def num_training_epochs(self) -> int:
        """Total training epochs inferred from datamodule and devices."""
        return self.trainer.max_epochs

    @property
    @lru_cache(maxsize=2)  # the size should always == 1
    def num_training_steps(self) -> int:
        """Total training steps inferred from datamodule and devices."""
        if self.trainer.max_steps != -1:
            return self.trainer.max_steps
        # FIXED: https://github.com/PyTorchLightning/pytorch-lightning/pull/11599
        # NEED TEST
        return (
            int(self.trainer.estimated_stepping_batches)
            // self.trainer.accumulate_grad_batches
        )

    def configure_optimizers(self):
        """Configure the optimizer and scheduler"""

        self.calc_loss.update_config(
            rt_config={
                "total_steps": self.num_training_steps,
                "total_epochs": self.num_training_epochs,
            }
        )

        optimizer_configs = self.config._get_modules("optimizer")
        assert (
            len(optimizer_configs) == 1
        ), "The optimizer submodule should only have one config"
        optimizer = register.get(
            "optimizer", register_module_name(optimizer_configs[0]._module_name)
        )(self.model, optimizer_configs[0]).get_optimizer()

        scheduler_configs = self.config._get_modules("scheduler")
        assert (
            len(scheduler_configs) == 1
        ), f"The scheduler submodule should only have one config. {scheduler_configs}"
        scheduler = register.get(
            "scheduler", register_module_name(scheduler_configs[0]._module_name)
        )(
            optimizer,
            scheduler_configs[0],
            rt_config={
                "num_training_steps": self.num_training_steps,
                "num_training_epochs": self.num_training_epochs,
            },
        ).get_scheduler()
        return {
            "optimizer": optimizer,
            "lr_scheduler": {
                "scheduler": scheduler,
                "interval": "step",
                "frequency": 1,
            },
        }



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Dict, List

import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from lightning.pytorch.callbacks import EarlyStopping

from dlk.utils.register import register


@cregister("callback", "early_stop")
class EarlyStoppingCallbackConfig(Base):
    """callback for the early stopping"""

    monitor = StrField(value=MISSING, help="the monitor metrics for checkpoint")
    mode = StrField(
        value="max",
        options=["min", "max"],
        help="max or min for monitor metrics",
    )
    patience = IntField(value=3, minimum=0, help="patience times for early stop")
    min_delta = FloatField(
        value=0.0,
        help="trigger early stop when the monitor metrics change less than min_delta",
    )
    check_on_train_epoch_end = BoolField(
        value=False,
        help="whether to run early stopping at the end of the training epoch.If this is ``False``, then the check runs at the end of the validation.",
    )
    strict = BoolField(
        value=True,
        help="whether to crash the training if `monitor` is not found in the validation metrics.",
    )
    stopping_threshold = FloatField(
        value=None,
        additions=[None],
        help="Stop training immediately once the monitored quantity reaches this threshold.",
    )
    divergence_threshold = FloatField(
        value=None,
        additions=[None],
        help="Stop training as soon as the monitored quantity becomes worse than this threshold.",
    )
    verbose = BoolField(value=True, help="verbosity mode.")
    log_rank_zero_only = BoolField(
        value=True,
        help="When set ``True``, logs the status of the early stopping callback only for rank 0 process.",
    )


@register("callback", "early_stop")
class EarlyStoppingCallback(object):
    """Early stop decided by config"""

    def __init__(self, config: EarlyStoppingCallbackConfig):
        super().__init__()
        self.config = config._to_dict(only_para=True)

    def __call__(self, rt_config: Dict) -> EarlyStopping:
        """return EarlyStopping object

        Args:
            rt_config: runtime config, include save_dir, and the checkpoint path name

        Returns:
            EarlyStopping object

        """
        return EarlyStopping(**self.config)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register


@cregister("initmethod", "range_uniform")
class RangeUniformInitConfig:

    """
    init the parameters by the uniform distribution with the given range
    """

    range_from = FloatField(value=-0.1, help="the lower bound of the init value")
    range_to = FloatField(value=0.1, help="the upper bound of the init value")


@register("initmethod", "range_uniform")
class RangeUniformInit(object):
    """for transformers"""

    def __init__(self, config: RangeUniformInitConfig):
        super().__init__()
        self.config = config

    def __call__(self, module):
        """Initialize the weights"""
        if isinstance(module, nn.Linear):
            module.weight.data.uniform_(
                from_=self.config.range_from, to=self.config.range_to
            )
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.uniform_(
                from_=self.config.range_from, to=self.config.range_to
            )
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            # use the default init,
            # weight are initialized to 1, bias to 0
            module.reset_parameters()
        elif isinstance(module, nn.Conv1d):
            module.weight.data.uniform_(
                from_=self.config.range_from, to=self.config.range_to
            )
        elif isinstance(module, nn.Conv2d):
            module.weight.data.uniform_(
                from_=self.config.range_from, to=self.config.range_to
            )
        elif isinstance(module, nn.Conv3d):
            module.weight.data.uniform_(
                from_=self.config.range_from, to=self.config.range_to
            )



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any, Callable, Dict, List, Union

import pandas as pd

from dlk.online import OnlinePredict
from dlk.preprocess import PreProcessor

logger = logging.getLogger(__name__)


class Server(object):
    """Demo"""

    def __init__(
        self,
        process_config: Union[str, Dict] = "/path/to/config",
        fit_config: Union[str, Dict] = "/path/to/config",
        checkpoint: str = "/path/to/checkpoint",
    ):
        super(Server, self).__init__()
        self.processor = self.get_processor(process_config)
        self.predictor = self.get_predictor(fit_config, checkpoint)

    def fit(self, input_df: pd.DataFrame) -> Dict:
        """use the model to get the result of the data want to predict

        Args:
            input_df: the one input data, it could be only one row or multiple rows(batch)

        Returns:
            the prediction result

        """
        processed_data = self.processor.fit(input_df)
        result = self.predictor.predict.predict(processed_data)
        return result

    def get_processor(self, config: Union[str, Dict]) -> PreProcessor:
        """get the preprocessor instance

        Args:
            config: the path to config or the confic dict

        Returns:
            PreProcessor instance

        """
        return PreProcessor(config, stage="online")

    def get_predictor(self, config: Union[str, Dict], checkpoint) -> OnlinePredict:
        """get the preprocessor instance

        Args:
            config: the path to config or the confic dict

        Returns:
            PreProcessor instance

        """
        return OnlinePredict(config, checkpoint)



# Copyright the author(s) of DLK.

# There are many code copy from fairseq.
# Copyright (c) Facebook, Inc. and its affiliates.


from typing import Dict, List, Optional

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch import Tensor

from dlk.utils.register import register

from . import Sample
from .beam_search import BeamSearch


@cregister("token_sample", "diverse_beam_search")
class DiverseBeamSearchConfig(Base):
    num_groups = IntField(value="???", help="the number of diverse groups")
    diversity_strength = FloatField(
        value=0.5,
        help="the strength of diversity, from 0.2 to 0.8 work well for most tasks( https://arxiv.org/pdf/1610.02424.pdf",
    )


@register("token_sample", "diverse_beam_search")
class DiverseBeamSearch(Sample):
    """Diverse Beam Search.

    See "Diverse Beam Search: Decoding Diverse Solutions from Neural Sequence
    Models" for details.

    We only implement the Hamming Diversity penalty here, which performed best
    in the original paper.
    """

    def __init__(self, tgt_dict, config: DiverseBeamSearchConfig):
        super().__init__(tgt_dict)
        self.num_groups = config.num_groups
        self.diversity_strength = -config.diversity_strength
        self.beam = BeamSearch(tgt_dict)

    def step(
        self,
        step: int,
        lprobs,
        scores,
        prev_output_tokens: Optional[Tensor] = None,
        original_batch_idxs: Optional[Tensor] = None,
    ):
        bsz, beam_size, vocab_size = lprobs.size()
        if beam_size % self.num_groups != 0:
            raise ValueError(
                "DiverseBeamSearch requires --beam to be divisible by the number of groups"
            )

        # initialize diversity penalty
        diversity_buf = torch.zeros(lprobs[:, 0, :].size()).to(lprobs)

        scores_G, indices_G, beams_G = [], [], []
        for g in range(self.num_groups):
            lprobs_g = lprobs[:, g :: self.num_groups, :]
            scores_g = scores[:, g :: self.num_groups, :] if step > 0 else None

            # apply diversity penalty
            if g > 0:
                lprobs_g = torch.add(
                    lprobs_g,
                    other=diversity_buf.unsqueeze(1),
                    alpha=self.diversity_strength,
                )
            else:
                lprobs_g = lprobs_g.contiguous()

            scores_buf, indices_buf, beams_buf = self.beam.step(
                step, lprobs_g, scores_g
            )
            beams_buf.mul_(self.num_groups).add_(g)

            scores_G.append(scores_buf.clone())
            indices_G.append(indices_buf.clone())
            beams_G.append(beams_buf.clone())

            # update diversity penalty
            diversity_buf.scatter_add_(
                1, indices_buf, torch.ones(indices_buf.size()).to(diversity_buf)
            )

        # interleave results from different groups
        scores_buf = torch.stack(scores_G, dim=2).view(bsz, -1)
        indices_buf = torch.stack(indices_G, dim=2).view(bsz, -1)
        beams_buf = torch.stack(beams_G, dim=2).view(bsz, -1)
        return scores_buf, indices_buf, beams_buf



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Set

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import SimpleModule
from dlk.nn.module.module_vit import Vit, VitConfig
from dlk.utils.register import register


@cregister("embedding", "vit")
class VitEmbeddingConfig(VitConfig):
    """
    the bert like embedding module
    """

    embedding_dim = IntField(value=MISSING, minimum=1, help="the embedding dim")

    class InputMap:
        pixel_values = StrField(value="pixel_values", help="the input ids")

    input_map = NestField(
        value=InputMap, help="the input map of the static embedding module"
    )

    class OutputMap:
        embedding = StrField(value="embedding", help="the embedding")

    output_map = NestField(
        value=OutputMap, help="the output map of the static embedding module"
    )


@register("embedding", "vit")
class VitEmbedding(SimpleModule):
    """Wrap the hugingface transformers"""

    def __init__(self, config: VitEmbeddingConfig):
        super(VitEmbedding, self).__init__(config)
        self.config = config
        self.pretrained_transformers = Vit(config)

    def init_weight(self, method):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        self.pretrained_transformers.init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """get the transformers output as embedding

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        pixel_values = inputs[self.config.input_map.pixel_values]
        (
            sequence_output,
            all_hidden_states,
            all_self_attentions,
        ) = self.pretrained_transformers({"pixel_values": pixel_values})
        inputs[self.config.output_map.embedding] = sequence_output
        return inputs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import abc
import importlib
import os
from typing import Callable, Dict, Type

import pandas as pd
import pyarrow.parquet as pq
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    dataclass,
)

from dlk.utils.import_module import import_module_dir


@dataclass
class BaseSubProcessorConfig(Base):
    """the base subprocessor"""

    train_data_set = ListField(
        value=[],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )
    predict_data_set = ListField(
        value=[],
        suggestions=[["predict"], []],
        help="the data set should be processed for predict stage, only predict data set will be processed or none of the data set will be processed",
    )
    online_data_set = ListField(
        value=[],
        suggestions=[["online"], []],
        help="the data set should be processed for online stage, only online data set will be processed or none of the data set will be processed",
    )
    input_map = DictField(
        value={},
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )
    output_map = DictField(
        value={},
        help="the output map of the processor, the key is the name of the processor provided key, the value is the nexted processor needed key",
    )


class BaseSubProcessor(object):
    """docstring for ISubProcessor"""

    def __init__(self, stage: str, config: BaseSubProcessorConfig, meta_dir: str):
        self.loaded_meta = False
        self.meta_dir = meta_dir

    def load_meta(self):
        self.loaded_meta = True

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """SubProcess entry

        Args:
            data:
            >>> |sentence |label|
            >>> |---------|-----|
            >>> |sent_a...|la   |
            >>> |sent_b...|lb   |

            deliver_meta:
                if there are some meta info need to deliver to next processor, and deliver_meta is True, save the meta info to datadir
        Returns:
            processed data

        """
        raise NotImplementedError


# automatically import any Python files in the models directory
subprocessor_dir = os.path.dirname(__file__)
import_module_dir(subprocessor_dir, "dlk.data.subprocessor")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

"""
Get the dlk package root path
"""
import os

ROOT = os.path.join("/")


def get_root():
    """get the dlk root

    Returns:
        abspath of this package

    """
    cur_dir = os.path.dirname(os.path.abspath(__file__))

    def check_cur_is_root(cur_dir):
        """check cur_dir has 'dlk' or not"""
        root_sign_list = ["dlk"]
        for root_sign in root_sign_list:
            if os.path.exists(os.path.join(cur_dir, root_sign)):
                return True
        return False

    while cur_dir != ROOT:
        if check_cur_is_root(cur_dir):
            break
        cur_dir = os.path.dirname(cur_dir)

    if cur_dir == ROOT:
        raise ImportError(
            f"{os.path.dirname(os.path.abspath(__file__))} is not a right file to find dlk root"
        )
    else:
        return cur_dir



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Union

from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from lightning.pytorch.callbacks import Callback, StochasticWeightAveraging

from dlk.utils.register import register


@cregister("callback", "weight_average")
class StochasticWeightAveragingCallbackConfig(Base):
    """callback for the Stochastic Weight Averaging to average a model weight on the training process"""

    use_swa = BoolField(
        value=True,
        help="whether use swa, effective on the imodel, if the config include swa and not set use_swa to False, default is True",
    )

    swa_lrs = AnyField(
        value=None,
        help="The SWA learning rate to use: float. Use this value for all parameter groups of the optimizer. List[float]. A list values for each parameter group of the optimizer.",
    )
    swa_epoch_start = FloatField(
        value=0.8,
        minimum=0.0,
        maximum=1.0,
        help="If provided as int, the procedure will start from the ``swa_epoch_start``-th epoch. If provided as float between 0 and 1, the procedure will start from ``int(swa_epoch_start * max_epochs)`` epoch",
    )
    annealing_epochs = IntField(
        value=10,
        minimum=0,
        help="""number of epochs in the annealing phase (default: 10)""",
    )
    annealing_strategy = StrField(
        value="cos",
        options=["cos", "linear"],
        help="""Specifies the annealing strategy (default: "cos"): ``"cos"``. For cosine annealing. ``"linear"`` For linear annealing""",
    )
    device = StrField(
        value="cpu",
        options=["cpu", "cuda"],
        help="""if provided, the averaged model will be stored on the ``device``. When None is provided, it will infer the `device` from ``pl_module``. (default: ``"cpu"``)""",
    )


@register("callback", "weight_average")
class StochasticWeightAveragingCallback(object):
    """Average weight by config"""

    def __init__(self, config: StochasticWeightAveragingCallbackConfig):
        super().__init__()
        self.config = config._to_dict(only_para=True)
        self.use_swa = self.config.pop("use_swa")

    def __call__(self, rt_config: Dict) -> Union[StochasticWeightAveraging, Callback]:
        """return StochasticWeightAveraging object

        Args:
            rt_config: runtime config, include save_dir, and the checkpoint path name

        Returns:
            StochasticWeightAveraging object

        """
        if not self.use_swa:
            return Callback()

        return StochasticWeightAveraging(**self.config)



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
from functools import partial
from typing import Callable, Dict, List, Set

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from tokenizers import Tokenizer

from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

from . import BaseSubProcessor, BaseSubProcessorConfig

logger = logging.getLogger(__name__)


@cregister("subprocessor", "token_embedding")
class TokenEmbeddingConfig(BaseSubProcessorConfig):
    """the token embedding subprocessor"""

    train_data_set = ListField(
        value=["train"],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )
    embedding_file = StrField(
        value=MISSING,
        suggestions=["embedding_file"],
        help="the embedding file path",
    )
    tokenizer_path = StrField(
        value=None,
        suggestions=["tokenizer"],
        additions=[None],
        help="the tokenizer path, when this is provided, we will use the tokenizer to get the vocab, not effective by `meta_dir`",
    )
    vocab = StrField(
        value=None,
        suggestions=["vocab"],
        additions=[None],
        help="the vocab path(dump by `Vocabulary`), when this is provided",
    )
    token_embedding = StrField(
        value="token_embedding",
        suggestions=["token_embedding"],
        help="the embedding saved name",
    )
    embedding_size = IntField(value=MISSING, minimum=1, help="the embedding size")

    class BiasClipRange:
        lower = FloatField(
            value=-10e9,
            help="the init embedding bias weight range lower bound",
        )
        upper = FloatField(
            value=10e9,
            help="the init embedding bias weight range upper bound",
        )

    bias_clip_range = NestField(
        value=BiasClipRange, help="the init embedding bias weight range"
    )


@register("subprocessor", "token_embedding")
class TokenEmbedding(BaseSubProcessor):
    """
    Gather tokens embedding from pretrained 'embedding_file' or init embedding(xavier_uniform init, and the range clip in 'bias_clip_range')

    The tokens are from 'Tokenizer'(get_vocab) or 'Vocabulary'(word2idx) object(the two must provide only one)
    """

    def __init__(self, stage: str, config: TokenEmbeddingConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.stage = stage
        self.config = config
        self.tokenizer: Tokenizer = None
        self.vocab: Vocabulary = None
        if self.config.embedding_file:
            self.origin_embedding = self.get_embedding(
                self.config.embedding_file, self.config.embedding_size
            )
        else:
            self.origin_embedding = {}

    def load_meta(self):
        self.loaded_meta = True
        if self.config.vocab:
            self.vocab = Vocabulary.load_from_file(
                os.path.join(self.meta_dir, self.config.vocab)
            )
        if self.config.tokenizer_path:
            with open(self.config.tokenizer_path, "r", encoding="utf-8") as f:
                tokenizer_str = json.dumps(json.load(f))
            self.tokenizer = Tokenizer.from_str(tokenizer_str)

    def get_embedding(self, file_path, embedding_size) -> Dict[str, List[float]]:
        """load the embeddings from file_path, and only get the last embedding_size dimensions embedding

        Args:
            file_path: embedding file path
            embedding_size: the embedding dim

        Returns:
            >>> embedding_dict
            >>> {
            >>>     "word": [embedding, ...]
            >>> }
        """
        embedding_dict = {}
        with open(file_path, "r", encoding="utf-8") as f:
            lines = f.readlines()
            for i, line in enumerate(lines):
                # if the first line is statistic info, continue
                if i == 0 and len(line.split()) <= embedding_size:
                    continue
                sp_line = line.split()
                if len(sp_line) <= embedding_size:
                    logger.warning(
                        f"The {i}th line len： {len(sp_line)}, token is {sp_line[0]}"
                    )
                    continue
                word = sp_line[0]
                vector = list(map(float, sp_line[-embedding_size:]))
                embedding_dict[word] = vector
        return embedding_dict

    def update_embedding(self, embedding_dict: Dict[str, List[float]], vocab: Dict):
        """update the embedding_dict which token in vocab but not in embedding_dict

        Args:
            embedding_dict: word->embedding dict
            vocab: token vocab

        Returns:
            updated embedding_dict

        """
        without_embedding_tokens = 0
        fuzzy_match_tokens = 0
        low: float = self.config.bias_clip_range.lower
        up: float = self.config.bias_clip_range.upper
        # xavier_uniform_ init method
        bias: float = np.sqrt(6.0 / (len(vocab) + self.config.embedding_size))
        if bias > up:
            bias = up
        elif bias < low:
            bias = low
            # bias = np.sqrt(3/self.config.embedding_size)
        for token in vocab:
            if token not in embedding_dict:
                if (token.lower() not in embedding_dict) and (
                    token.upper() not in embedding_dict
                ):
                    embedding_dict[token] = list(
                        np.random.uniform(-bias, bias, self.config.embedding_size)
                    )
                    without_embedding_tokens += 1
                else:
                    fuzzy_match_tokens += 1
                    if token.lower() in embedding_dict:
                        embedding_dict[token] = embedding_dict[token.lower()]
                    else:
                        embedding_dict[token] = embedding_dict[token.upper()]
        logger.info(
            f"All tokens num is {len(vocab)}, fuzzy matching(lower or upper match) num is {fuzzy_match_tokens}, OOV token num is {without_embedding_tokens}"
        )
        return embedding_dict

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """Character gather entry

        Args:
            data:
            >>> |sentence |label|
            >>> |---------|-----|
            >>> |sent_a...|la   |
            >>> |sent_b...|lb   |

            deliver_meta:
                if there are some meta info need to deliver to next processor, and deliver_meta is True, save the meta info to datadir
        Returns:
            processed data

        """
        if not deliver_meta:
            return data

        if not self.loaded_meta:
            self.load_meta()

        if self.tokenizer is not None and self.vocab:
            raise PermissionError(f"The tokenizer and vocab must provide one.")
        if self.tokenizer:
            token2id: Dict = self.tokenizer.get_vocab()
            id2token: Dict = {value: key for key, value in token2id.items()}
        else:
            token2id: Dict = self.vocab.word2idx
            id2token: Dict = self.vocab.idx2word
        embedding_dict = self.update_embedding(self.origin_embedding, token2id)
        embedding_mat = np.array(
            [embedding_dict[id2token[id]] for id in range(len(id2token))]
        )

        with open(os.path.join(self.meta_dir, self.config.token_embedding), "wb") as f:
            embedding_mat.dump(f)

        return data



# Copyright the author(s) of DLK.

# There are many code copy from fairseq.
# Copyright (c) Facebook, Inc. and its affiliates.


import math
from typing import Dict, List, Optional

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch import Tensor

from dlk.utils.register import register

from . import Sample


@cregister("token_sample", "prefix_constrained_beam_search")
class PrefixConstrainedBeamSearchConfig(Base):
    gen_prefixs_fn = StrField(
        value="???",
        help="the prefix allowed tokens function, registerd in gen_prefixs_fn",
    )


@register("token_sample", "prefix_constrained_beam_search")
class PrefixConstrainedBeamSearch(Sample):
    def __init__(self, tgt_dict, config: PrefixConstrainedBeamSearchConfig):
        super().__init__(tgt_dict)
        self.gen_prefixs_fn = register.get("gen_prefixs_fn", config.gen_prefixs_fn)
        self.stop_on_max_len = True

    def apply_mask(self, x, prev_output_tokens, original_batch_idxs):
        beam_size = x.shape[0] // original_batch_idxs.shape[0]
        original_batch_idxs = (
            original_batch_idxs.unsqueeze(-1).repeat((1, beam_size)).flatten().tolist()
        )

        mask = torch.full_like(x, -math.inf)
        for sent_i, (sent, batch_i) in enumerate(
            zip(prev_output_tokens, original_batch_idxs)
        ):
            mask[sent_i, :, self.gen_prefixs_fn(batch_i, sent)] = 0

        return mask

    def step(
        self,
        step: int,
        lprobs: Tensor,
        scores: Tensor,
        prev_output_tokens: Tensor,
        original_batch_idxs: Tensor,
    ):
        bsz, beam_size, vocab_size = lprobs.size()

        lprobs += self.apply_mask(
            lprobs.view(bsz * beam_size, 1, vocab_size),
            prev_output_tokens,
            original_batch_idxs,
        ).view(bsz, beam_size, vocab_size)

        if step == 0:
            # at the first step all hypotheses are equally likely, so use
            # only the first beam
            lprobs = lprobs[:, ::beam_size, :].contiguous()
        else:
            # make probs contain cumulative scores for each hypothesis
            assert scores is not None
            lprobs = lprobs + scores[:, :, step - 1].unsqueeze(-1)

        top_prediction = torch.topk(
            lprobs.view(bsz, -1),
            k=min(
                # Take the best beam_size predictions. We'll choose the first
                # beam_size of these which don't predict eos to continue with.
                beam_size,
                lprobs.view(bsz, -1).size(1) - 1,  # -1 so we never select pad
            ),
        )
        scores_buf = top_prediction[0]
        indices_buf = top_prediction[1]
        beams_buf = indices_buf // vocab_size
        indices_buf = indices_buf.fmod(vocab_size)
        return scores_buf, indices_buf, beams_buf



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Dict

from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    dataclass,
)
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR

from dlk.utils.import_module import import_module_dir


@dataclass
class BaseSchedulerConfig(Base):
    """the base scheduler"""

    num_warmup_steps = FloatField(
        value=0.1,
        minimum=0.0,
        help="""
            only for the scheduler with warmup.
            for the interval is `step`,
            the number of warmup steps,
            it can be float or int, 
            if it's between 0.0-1.0, means the percentage of total steps,
            if it's larger than 1, means the number of steps.
            If the interval is `epoch`,
            it means the number of warmup epochs
        """,
    )
    last_epoch = IntField(value=-1, minimum=-1, help="the last epoch/step")
    interval = StrField(
        value="step", options=["epoch", "step"], help="the interval of scheduler"
    )


class BaseScheduler(object):
    """interface for Schedule"""

    def __init__(
        self, optimizer: Optimizer, config: BaseSchedulerConfig, rt_config: Dict
    ):
        """update dynamic parameters in self.config based the rt_config
        Args:
            rt_config: provide the current training status
                >>> {
                >>>     "num_training_steps": self.num_training_steps,
                >>>     "num_training_epochs": self.num_training_epochs
                >>> }
        Returns: None
        """
        super(BaseScheduler, self).__init__()
        self.config = config
        self.optimizer = optimizer
        self.num_training_steps = rt_config["num_training_steps"]
        self.num_training_epochs = rt_config["num_training_epochs"]
        if self.config.num_warmup_steps < 1:
            if self.config.interval == "step":
                self.config.num_warmup_steps = int(
                    self.config.num_warmup_steps * self.num_training_steps
                )
            else:
                assert self.config.interval == "epoch"
                self.config.num_warmup_steps = int(
                    self.config.num_warmup_steps * self.num_training_epochs
                )

    def get_scheduler(self) -> LambdaLR:
        """return the initialized scheduler

        Returns:
            Schedule
        """
        if self.config.interval == "step":
            return LambdaLR(
                self.optimizer, self.step_update, last_epoch=self.config.last_epoch
            )
        else:
            assert self.config.interval == "epoch"
            return LambdaLR(
                self.optimizer, self.epoch_update, last_epoch=self.config.last_epoch
            )

    def step_update(self, current_step: int):
        raise NotImplementedError

    def epoch_update(self, current_epoch: int):
        raise NotImplementedError

    def __call__(self):
        """the same as self.get_scheduler()"""
        return self.get_scheduler()


scheduler_dir = os.path.dirname(__file__)
import_module_dir(scheduler_dir, "dlk.scheduler")



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from typing import Callable, Dict, List, Set

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

from . import BaseSubProcessor, BaseSubProcessorConfig

logger = logging.getLogger(__name__)


@cregister("subprocessor", "span_cls_relabel")
class SpanClsRelabelConfig(BaseSubProcessorConfig):
    """the span classification relabel subprocessor"""

    train_data_set = ListField(
        value=["train", "valid", "test"],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )

    class InputMap:
        word_ids = StrField(value="word_ids", help="the word ids")
        offsets = StrField(value="offsets", help="the offsets")
        entities_info = StrField(value="entities_info", help="the entities info")

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OutputMap:
        label_ids = StrField(value="label_ids", help="the label ids")
        processed_entities_info = StrField(
            value="processed_entities_info", help="the processed entities info"
        )

    output_map = NestField(
        value=OutputMap,
        help="the output map of the processor, the key is the name of the processor provided key, the value is the nexted processor needed key",
    )
    drop = StrField(
        value="shorter",
        options=["longer", "shorter", "none"],
        help="the drop strategy for the overlap entities",
    )
    vocab = StrField(value="label_vocab.json", help="the vocab for the label")

    entity_priority = ListField(
        value=[],
        suggestions=[["Product", "Brand"]],
        help="the entity priority, when conflict, will keep the entity with the highest priority",
    )
    priority_trigger = IntField(
        value=1,
        help="if the overlap entity abs(length_a - length_b)<=priority_trigger, will trigger the entity_priority strategy",
    )
    mask_fill = IntField(value=-100, help="the mask fill for the label")
    mask_first_sent = BoolField(
        value=False, help="whether mask the first sentence for anwering"
    )
    null_to_zero_index = BoolField(
        value=False,
        help="if cannot find the entity, set to point to the first(zero) index token",
    )
    strict = BoolField(
        value=True, help="if strict == True, will drop the invalid sample"
    )


@register("subprocessor", "span_cls_relabel")
class SpanClsRelabel(BaseSubProcessor):
    """
    Relabel the char level entity span to token level and construct matrix
    """

    def __init__(self, stage: str, config: SpanClsRelabelConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.stage = stage
        self.config = config
        self.vocab: Vocabulary = None
        self.entity_priority = {
            entity: priority
            for priority, entity in enumerate(self.config.entity_priority)
        }

    def load_meta(self):
        self.loaded_meta = True
        self.vocab = Vocabulary.load_from_file(
            os.path.join(self.meta_dir, self.config.vocab)
        )
        assert (
            self.vocab.word2idx[self.vocab.unknown] == 0
        ), f"For span_cls_relabel, 'unknown' must be index 0, and other labels as 1...num_label"
        assert (
            not self.vocab.pad
        ), f"For span_cls_relabel, 'pad' must be index 0, and other labels as 1...num_label"

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """firstpiece relabel the data

        Args:
            data: one is like

            >>> {
            >>>     "uuid": '**-**-**-**'
            >>>     "sentence": "Mie Merah - Buah Bit",
            >>>     "offsets": see the offsets in fast_tokenizer
            >>>     "entities_info": [
            >>>                 {
            >>>                     "end": 9,
            >>>                     "start": 0,
            >>>                     "labels": [
            >>>                         "Product"
            >>>                     ]
            >>>                 },
            >>>             ]
            >>>         },
            >>>     ],
            >>> },

            deliver_meta:
                ignore
        Returns:
            relabeld data
        """
        if not self.loaded_meta:
            self.load_meta()

        data[
            [
                self.config.output_map.label_ids,
                self.config.output_map.processed_entities_info,
            ]
        ] = data.apply(self.relabel, axis=1, result_type="expand")
        if self.config.strict:
            data.dropna(axis=0, inplace=True)
            data.reset_index(inplace=True)

        return data

    def find_position_in_offsets(
        self,
        position: int,
        offset_list: List,
        sub_word_ids: List,
        start: int,
        end: int,
        is_start: bool = False,
    ):
        """find the sub_word index which the offset_list[index][0]<=position<offset_list[index][1]

        Args:
            position: position
            offset_list: list of all tokens offsets
            sub_word_ids: word_ids from tokenizer
            start: start search index
            end: end search index
            is_start: is the position is the start of target token, if the is_start==True and cannot find return -1

        Returns:
            the index of the offset which include position

        """
        while start < end:
            if sub_word_ids[start] is None:
                start += 1
            elif position >= offset_list[start][0] and position < offset_list[start][1]:
                return start
            elif position < offset_list[start][0]:
                if start == 1 and offset_list[0] == [0, 0]:
                    return 1
                if is_start:
                    return -1
                else:
                    return start - 1
            else:
                start += 1
        return -1

    def relabel(self, one_ins: pd.Series):
        """make token label, if use the first piece label please use the 'span_cls_firstpiece_relabel'

        Args:
            one_ins: include sentence, entity_info, offsets

        Returns:
            labels(labels for each subtoken)
            entities_info
            processed_entities_info: for relation relabal
        """
        pre_clean_entities_info: List = one_ins[self.config.input_map.entities_info]
        if self.config.drop != "none":
            pre_clean_entities_info.sort(key=lambda x: x["start"])
        offsets: List = one_ins[self.config.input_map.offsets]
        sub_word_ids: List = one_ins[self.config.input_map.word_ids]

        if self.config.mask_first_sent:
            first_start = sub_word_ids.index(0)
            second_start = sub_word_ids[first_start + 1 :].index(0)
            mask_first_index = first_start + second_start + 2
        else:
            mask_first_index = 0  # if there is only one sentence, set to 0

        entities_info = []
        pre_end = -1
        pre_length = 0
        pre_label = ""
        for entity_info in pre_clean_entities_info:
            assert (
                len(entity_info["labels"]) == 1
            ), f"currently we just support one label for one entity"
            if entity_info["start"] < pre_end:  # if overlap will remove one
                if self.config.drop == "none":
                    pass
                elif (
                    abs(entity_info["end"] - entity_info["start"] - pre_length)
                    <= self.config.priority_trigger
                ):
                    pre_label_order = self.entity_priority.get(pre_label, 1e9)
                    label_order = self.entity_priority.get(
                        entity_info["labels"][0], 1e9
                    )
                    if label_order < pre_label_order:
                        entities_info.pop()
                    else:
                        continue
                elif self.config.drop == "shorter":
                    if entity_info["end"] - entity_info["start"] > pre_length:
                        entities_info.pop()
                    else:
                        continue
                elif self.config.drop == "longer":
                    if entity_info["end"] - entity_info["start"] < pre_length:
                        entities_info.pop()
                    else:
                        continue
                else:
                    raise PermissionError(
                        f"The drop method must in 'none'/'shorter'/'longer'"
                    )
            pre_label = entity_info["labels"][0]
            entities_info.append(entity_info)
            pre_end = entity_info["end"]
            pre_length = entity_info["end"] - entity_info["start"]

        offset_length = len(offsets)

        unknown_id = self.vocab.get_index(self.vocab.unknown)
        mask_matrices = np.full(
            (offset_length, offset_length), self.config.mask_fill, dtype=np.int8
        )
        mask_matrices = np.tril(mask_matrices, k=-1)

        unknown_matrices = np.full(
            (offset_length, offset_length), unknown_id, dtype=np.int8
        )
        unknown_matrices = np.triu(unknown_matrices, k=0)

        label_matrices = unknown_matrices + mask_matrices
        if sub_word_ids[0] is None:
            label_matrices[0, :] = self.config.mask_fill
        if sub_word_ids[-1] is None:
            label_matrices[:, -1] = self.config.mask_fill

        if mask_first_index > 0:
            label_matrices[:mask_first_index, :] = self.config.mask_fill
            label_matrices[:, :mask_first_index] = self.config.mask_fill
        processed_entities_info = []
        for entity_info in entities_info:
            if entity_info["start"] == 0 and entity_info["end"] == 0:
                start_token_index, end_token_index = 0, 0
            else:
                start_token_index = self.find_position_in_offsets(
                    entity_info["start"],
                    offsets,
                    sub_word_ids,
                    mask_first_index,
                    offset_length,
                    is_start=True,
                )
                if start_token_index == -1:
                    if self.config.null_to_zero_index:
                        start_token_index, end_token_index = 0, 0
                    else:
                        if self.config.strict:
                            logger.warning(
                                f"cannot find the entity_info : {entity_info}, offsets: {offsets}, we will drop this instance"
                            )
                            return None, None
                        logger.warning(
                            f"cannot find the entity_info : {entity_info}, offsets: {offsets}"
                        )
                        continue
                else:
                    end_token_index = self.find_position_in_offsets(
                        entity_info["end"] - 1,
                        offsets,
                        sub_word_ids,
                        start_token_index,
                        offset_length,
                    )
                    if self.config.null_to_zero_index and end_token_index == -1:
                        start_token_index, end_token_index = 0, 0
            assert (
                end_token_index != -1
            ), f"entity_info: {entity_info}, offsets: {offsets}"
            label_id = self.vocab.get_index(entity_info["labels"][0])
            label_matrices[start_token_index, end_token_index] = label_id
            entity_info["sub_token_start"] = start_token_index
            entity_info["sub_token_end"] = end_token_index
            processed_entities_info.append(entity_info)

        return label_matrices, processed_entities_info



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Callable, Dict, List, Set

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.io import open
from dlk.utils.register import register

from . import BaseSubProcessor, BaseSubProcessorConfig

logger = logging.getLogger(__name__)


@cregister("subprocessor", "piece_rerank_relabel")
class PieceRerankRelabelConfig(BaseSubProcessorConfig):
    """the piece rerank relabel subprocessor"""

    train_data_set = ListField(
        value=["train", "valid", "test"],
        suggestions=[["train", "valid", "test"], []],
        help="the data set should be processed for train stage",
    )

    class InputMap:
        word_ids = StrField(value="word_ids", help="the word ids")
        offsets = StrField(value="offsets", help="the offsets")
        rank_info = StrField(value="rank_info", help="the rank info")

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OutputMap:
        label_ids = StrField(value="label_ids", help="the label ids")

    output_map = NestField(
        value=OutputMap,
        help="the output map of the processor, the key is the name of the processor provided key, the value is the nexted processor needed key",
    )
    mask_fill = IntField(value=-100, help="the mask fill value")


@register("subprocessor", "piece_rerank_relabel")
class PieceRerankRelabel(BaseSubProcessor):
    """
    Relabel the piece rank construct matrix
    """

    def __init__(self, stage: str, config: PieceRerankRelabelConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.stage = stage
        self.config = config

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """Character gather entry

        Args:
            data: one is like
            >>> {
            >>>     "uuid": "uuid",
            >>>     "pretokenized_words": ['[PAD]', "BCD", "E", "A"],
            >>>     "rank_info": [0, 3, 1, 2]
            >>> }

            deliver_meta:
                ignore
        Returns:
            processed data
        """
        data[self.config.output_map.label_ids] = data.apply(self.relabel, axis=1)

        return data

    def relabel(self, one_ins: pd.Series):
        """make token label, if use the first piece label please use the 'piece_rerank_firstpiece_relabel'

        Args:
            one_ins: include sentence, rank_info

        Returns:
            label_matrix for real rank
        """
        word_ids: List = one_ins[self.config.input_map.word_ids]
        rank_info = one_ins[self.config.input_map.rank_info]
        seq_len = len(word_ids)
        label_matrix = np.full((seq_len, seq_len), 0, dtype=np.int8)
        if not word_ids[0]:
            label_matrix[0] = self.config.mask_fill
            label_matrix[:, 0] = self.config.mask_fill
        if not word_ids[-1]:
            label_matrix[-1] = self.config.mask_fill
            label_matrix[:, -1] = self.config.mask_fill
        # {
        #     "uuid": "uuid",
        #     "pretokenized_words": ['[PAD]', "BCD", "E", "A"],
        #     "rank_info": [0, 3, 1, 2]
        # }
        # ['[CLS]', '[PAD]', 'BC', '##D', 'E', 'A', '[SEP]']
        # [(0, 0), (0, 5), (0, 2), (2, 3), (0, 1), (0, 1), (0, 0)]
        # [None, 0, 1, 1, 2, 3, None]
        # [101, 0, 3823, 2137, 142, 138, 102]
        assert rank_info[0] == 0
        pre_position = None
        for index in rank_info:
            cur_position = word_ids.index(index)
            cur_word_id = word_ids[cur_position]
            if pre_position:
                label_matrix[pre_position][cur_position] = 1
            while (
                cur_position + 2 < seq_len and word_ids[cur_position + 1] == cur_word_id
            ):
                label_matrix[cur_position][cur_position + 1] = 1
                cur_position = cur_position + 1
            pre_position = cur_position
        start_position = word_ids.index(0)
        label_matrix[pre_position][start_position] = 1
        return label_matrix



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Dict

import torch
import torch.nn as nn
from intc import (
    AnyField,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import BaseIdentityModule
from dlk.utils.import_module import import_module_dir
from dlk.utils.register import register


@cregister("module", "identity")
class IdentityModuleConfig:
    """identity module"""

    pass


@register("module", "identity")
class IdentityModule(BaseIdentityModule):
    pass


class Module(nn.Module):
    """This class is means DLK Module for replace the torch.nn.Module in this project"""

    def init_weight(self, method):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        for module in self.children():
            module.apply(method)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """in simple module, all step fit to this method

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        raise NotImplementedError

    def predict_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do predict for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        return self(inputs)

    def training_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do train for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        return self(inputs)

    def validation_step(
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """do validation for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        return self(inputs)

    def test_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do test for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        return self(inputs)


module_dir = os.path.dirname(__file__)
import_module_dir(module_dir, "dlk.nn.module")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Collection, Dict, List

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.io import open
from dlk.utils.register import register

from . import Module


@cregister("module", "conv1d")
class Conv1dConfig:
    """the 1D convolution config"""

    in_channels = IntField(value=MISSING, minimum=0, help="the input channels")
    out_channels = IntField(value=MISSING, minimum=0, help="the output channels")
    dropout = FloatField(value=0.0, minimum=0.0, maximum=1.0, help="the dropout rate")
    kernel_sizes = ListField(value=[3], help="the kernel sizes")


@register("module", "conv1d")
class Conv1d(Module):
    """Conv for 1d input"""

    def __init__(self, config: Conv1dConfig):
        super().__init__()
        self.config = config
        assert all(
            k % 2 == 1 for k in self.config.kernel_sizes
        ), "the kernel sizes must be odd"
        assert (
            self.config.out_channels % len(self.config.kernel_sizes) == 0
        ), "out channels must be dividable by kernels"
        self.config.out_channels = self.config.out_channels // len(
            self.config.kernel_sizes
        )

        convs = []
        for kernel_size in self.config.kernel_sizes:
            conv = nn.Conv1d(
                self.config.in_channels,
                self.config.out_channels,
                kernel_size,
                padding=(kernel_size - 1) // 2,
            )
            convs.append(nn.Sequential(conv, nn.GELU()))
        self.convs = nn.ModuleList(convs)
        self.dropout = nn.Dropout(p=float(self.config.dropout))

    def forward(self, x: torch.Tensor):
        """do forward on a mini batch

        Args:
            batch: a mini batch inputs

        Returns:
            conv result the shape is the same as input

        """
        return self.dropout(torch.cat([conv(x) for conv in self.convs], dim=-1))



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
from typing import Callable, Dict, List, Set

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

from . import BaseSubProcessor, BaseSubProcessorConfig

logger = logging.getLogger(__name__)


@cregister("subprocessor", "span_relation_relabel")
class SpanRelationRelabelConfig(BaseSubProcessorConfig):
    """the span relation relabel subprocessor"""

    train_data_set = ListField(
        value=["train", "valid", "test"],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )

    class InputMap:
        word_ids = StrField(value="word_ids", help="the word ids")
        offsets = StrField(value="offsets", help="the offsets")
        relations_info = StrField(value="relations_info", help="the relations info")
        processed_entities_info = StrField(
            value="processed_entities_info", help="the processed entities info"
        )

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OutputMap:
        label_ids = StrField(value="label_ids", help="the label ids")

    output_map = NestField(
        value=OutputMap,
        help="the output map of the processor, the key is the name of the processor provided key, the value is the nexted processor needed key",
    )
    vocab = StrField(
        value="label_vocab#relation.json", help="the vocab for the relation label"
    )

    mask_fill = IntField(value=-100, help="the mask fill for the label")
    label_seperate = BoolField(
        value=False,
        help="separate differences types of relations to different label matrix or use universal matrix(universal matrix may have conflict issue, like the 2 entities of the head of 2 different relations is same)",
    )
    sym = BoolField(
        value=True,
        help="whether the from entity and end entity can swap in relations( if sym==True, we can just calc the upper trim and set down trim as -100 to ignore)",
    )
    strict = BoolField(
        value=True, help="if strict == True, will drop the invalid sample"
    )


@register("subprocessor", "span_relation_relabel")
class SpanRelationRelabel(BaseSubProcessor):
    """
    Relabel the json data to bio
    """

    def __init__(self, stage: str, config: SpanRelationRelabelConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.stage = stage
        self.config = config
        self.vocab: Vocabulary = None

    def load_meta(self):
        self.loaded_meta = True
        self.vocab = Vocabulary.load_from_file(
            os.path.join(self.meta_dir, self.config.vocab)
        )
        assert (
            self.vocab.word2idx[self.vocab.unknown] == 0
        ), f"For span_relation_relabel, 'unknown' must be index 0, and other labels as 1...num_label"
        assert (
            not self.vocab.pad
        ), f"For span_relation_relabel, 'pad' must be index 0, and other labels as 1...num_label"

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """firstpiece relabel the data

        Args:
            data: one is like

            >>> {
            >>>     "uuid": '**-**-**-**'
            >>>     "sentence": "Mie Merah - Buah Bit",
            >>>     "offsets": see the offsets in fast_tokenizer
            >>>     "processed_entities_info": generate by `span_cls_relabel`
            >>>     "relations_info": like
            >>>         [
            >>>             {
            >>>                 "labels": [
            >>>                     "belong_to"
            >>>                 ],
            >>>                 "from": 1, # index of entity
            >>>                 "to": 0,
            >>>             }
            >>>         ]
            >>> },

            deliver_meta:
                ignore
        Returns:
            relabeld data
        """
        if not self.loaded_meta:
            self.load_meta()

        data[self.config.output_map.label_ids] = data.apply(self.relabel, axis=1)
        if self.config.strict:
            data.dropna(axis=0, inplace=True)
            data.reset_index(inplace=True)

        return data

    def relabel(self, one_ins: pd.Series):
        """make token label, if use the first piece label please use the 'span_cls_firstpiece_relabel'

        Args:
            one_ins: include sentence, entity_info, offsets

        Returns:
            labels(labels for each subtoken)

        """
        relations_info = one_ins[self.config.input_map.relations_info]
        processed_entities_info = one_ins[self.config.input_map.processed_entities_info]
        entities_id_info_map = {}
        for processed_entity_info in processed_entities_info:
            entities_id_info_map[
                processed_entity_info["entity_id"]
            ] = processed_entity_info
        sub_word_ids = one_ins[self.config.input_map.word_ids]
        offsets = one_ins[self.config.input_map.offsets]
        offset_length = len(offsets)

        unknown_id = self.vocab.get_index(self.vocab.unknown)
        unknown_matrix = np.full((offset_length, offset_length), unknown_id)
        if self.config.sym:
            mask_matrix = np.full((offset_length, offset_length), self.config.mask_fill)
            mask_matrix = np.tril(mask_matrix, k=-1)
            unknown_matrix = np.triu(unknown_matrix, k=0)
            label_matrix = unknown_matrix + mask_matrix
        else:
            label_matrix = unknown_matrix
        if sub_word_ids[0] is None:
            label_matrix[0, :] = self.config.mask_fill
        if sub_word_ids[-1] is None:
            label_matrix[:, -1] = self.config.mask_fill
        if self.config.label_seperate:
            label_matrices = self.get_seperate_label_matrix(
                label_matrix, relations_info, entities_id_info_map
            )
        else:
            label_matrices = self.get_universal_label_matrix(
                label_matrix, relations_info, entities_id_info_map
            )

        return label_matrices

    def _get_entities_index(self, relation_info, entities_id_info_map):
        """get the from entity head, to entity head, from entity tail, to entity tail"""
        from_index = relation_info["from"]
        to_index = relation_info["to"]
        if (
            from_index not in entities_id_info_map
            or to_index not in entities_id_info_map
        ):
            return None, None, None, None
        from_entity = entities_id_info_map[from_index]
        from_start_index = from_entity["sub_token_start"]
        from_end_index = from_entity["sub_token_end"]
        to_entity = entities_id_info_map[to_index]
        to_start_index = to_entity["sub_token_start"]
        to_end_index = to_entity["sub_token_end"]

        if self.config.sym:
            if from_start_index > to_start_index:
                from_start_index, to_start_index = to_start_index, from_start_index
            if from_end_index > to_end_index:
                from_end_index, to_end_index = to_end_index, from_end_index
        return from_start_index, to_start_index, from_end_index, to_end_index

    def get_seperate_label_matrix(
        self, label_matrix, relations_info: List, entities_id_info_map: Dict[int, Dict]
    ):
        """separate different type of relations to different matrix(as one dimension)

        Args:
            label_matrices: matrix
            relations_info: like
                [
                    {
                        "labels": [
                            "belong_to"
                        ],
                        "from": 1, # index of entity
                        "to": 0,
                    }
                ]
            entities_id_info_map: generate by span_cls_relabel

        Returns:
            label_id matrix

        """
        label_types = len(self.vocab) - 1
        label_matrix = np.expand_dims(label_matrix, 0).repeat(2 * label_types, 0)

        for relation_info in relations_info:
            label_id = self.vocab.word2idx[relation_info["labels"][0]]
            (
                from_start_index,
                to_start_index,
                from_end_index,
                to_end_index,
            ) = self._get_entities_index(relation_info, entities_id_info_map)
            if from_start_index is None:
                if self.config.strict:
                    logger.warning(
                        f"Cannot get the relation of {json.dumps(relation_info)}, we will drop this instance"
                    )
                    return None
                continue
            # NOTE: label_id ==0 is unknown
            label_matrix[(label_id - 1) * 2][from_start_index][to_start_index] = 1
            label_matrix[(label_id - 1) * 2 + 1][from_end_index][to_end_index] = 1
        return label_matrix

    def get_universal_label_matrix(
        self, label_matrix, relations_info: List, entities_id_info_map: Dict[int, Dict]
    ):
        """different type of relations to one matrix

        Args:
            label_matrices: matrix
            relations_info: like
                [
                    {
                        "labels": [
                            "belong_to"
                        ],
                        "from": 1, # index of entity
                        "to": 0,
                    }
                ]
            entities_id_info_map: generate by span_cls_relabel

        Returns:
            label_id matrix
        """
        label_matrix = np.expand_dims(label_matrix, 0).repeat(2, 0)
        for relation_info in relations_info:
            label_id = self.vocab.word2idx[relation_info["labels"][0]]
            (
                from_start_index,
                to_start_index,
                from_end_index,
                to_end_index,
            ) = self._get_entities_index(relation_info, entities_id_info_map)
            if from_start_index is None:
                if self.config.strict:
                    logger.warning(
                        f"Cannot get the relation of {json.dumps(relation_info)}, we will drop this instance"
                    )
                    return None
                continue
            label_matrix[0][from_start_index][to_start_index] = label_id
            label_matrix[1][from_end_index][to_end_index] = label_id
        return label_matrix



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Any, Dict, List, Union

import pandas as pd
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.data.postprocessor import BasePostProcessor, BasePostProcessorConfig
from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

logger = logging.getLogger(__name__)


@cregister("postprocessor", "txt_reg")
class TxtRegPostProcessorConfig(BasePostProcessorConfig):
    """postprocess for text regression"""

    class InputMap:
        logits = StrField(value="logits", help="the output logits")
        value = StrField(value="values", help="the target value of the sample")
        index = StrField(value="_index", help="the index of the sample")

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OriginInputMap:
        uuid = StrField(value="uuid", help="the uuid or the id of the sample")
        sentence = StrField(
            value="sentence", help="the sentence of the sample(for data_type=single)"
        )
        sentence_a = StrField(
            value="sentence_a", help="the sentence_a of the sample(for data_type=pair)"
        )
        sentence_b = StrField(
            value="sentence_b", help="the sentence_b of the sample(for data_type=pair)"
        )

    origin_input_map = NestField(
        value=OriginInputMap,
        help="the origin input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )
    data_type = StrField(
        value="single", options=["single", "pair"], help="the data type, single or pair"
    )
    log_reg = BoolField(value=False, help="whether to return the log reg")


@register("postprocessor", "txt_reg")
class TxtRegPostProcessor(BasePostProcessor):
    """text regression postprocess"""

    def __init__(self, config: TxtRegPostProcessorConfig):
        super(TxtRegPostProcessor, self).__init__(config)
        self.config = config

    def do_predict(
        self,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> List:
        """Process the model predict to human readable format

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            all predicts
        """
        results = []
        for outputs in list_batch_outputs:
            logits = outputs[self.config.input_map.logits].detach()
            if self.config.log_reg:
                logits = torch.sigmoid(logits)
            assert len(logits.shape) == 2
            # predict_indexes = list(torch.argmax(logits, 1))
            indexes = list(outputs[self.config.input_map.index])

            if self.config.input_map.value in outputs:
                values = outputs[self.config.input_map.value]
            else:
                values = [0.0] * len(indexes)
            for i, (one_logits, index, value) in enumerate(
                zip(logits, indexes, values)
            ):
                one_ins = {}
                one_origin = origin_data.iloc[int(index)]
                if self.config.data_type == "single":
                    sentence = one_origin[self.config.origin_input_map.sentence]
                    one_ins["sentence"] = sentence
                else:
                    sentence_a = one_origin[self.config.origin_input_map.sentence_a]
                    one_ins["sentence_a"] = sentence_a
                    sentence_b = one_origin[self.config.origin_input_map.sentence_b]
                    one_ins["sentence_b"] = sentence_b

                uuid = one_origin[self.config.origin_input_map.uuid]
                one_ins["uuid"] = uuid
                one_ins["values"] = [float(value)]
                one_ins["predict_values"] = [float(one_logits)]
                one_ins["predict_extend_return"] = self.gather_predict_extend_data(
                    outputs, i, self.config.predict_extend_return
                )
                results.append(one_ins)
        return results

    def do_calc_metrics(
        self,
        predicts: List,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> Dict:
        """calc the scores use the predicts or list_batch_outputs

        Args:
            predicts: list of predicts

            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            the named scores, acc

        """
        ses = []
        for one_ins in predicts:
            values = one_ins["values"]
            assert (
                len(values) == 1
            ), "We currently is not support multi values in regression postprocess"
            value = values[0]

            one_predicts = one_ins["predict_values"]
            predict_value = one_predicts[0]
            ses.append((predict_value - value) ** 2)
        real_name = self.loss_name_map(stage)
        return {f"{real_name}_mse": sum(ses) / len(ses)}



# Copyright the author(s) of DLK.

# There are many code copy from fairseq.
# Copyright (c) Facebook, Inc. and its affiliates.


from typing import Dict, List, Optional

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch import Tensor

from dlk.utils.register import register

from . import Sample
from .beam_search import BeamSearch


@cregister("token_sample", "diverse_siblings_search")
class DiverseSiblingsSearchConfig(Base):
    # TODO: diversity_rate need test
    diversity_rate = FloatField(value=0.1, help="the diversity rate")


@register("token_sample", "diverse_siblings_search")
class DiverseSiblingsSearch(Sample):
    """
    Beam search with diverse siblings.

    See "A Simple, Fast Diverse Decoding Algorithm for Neural Generation" for details.
    https://arxiv.org/abs/1611.08562

    1/ Calculate hypotheses for each beam
    2/ Intra-sibling ordering
    3/ Rewrite scores
    4/ Choose top K hypotheses

    if diversity_rate == 0 is equivalent to BeamSearch
    """

    def __init__(self, tgt_dict, config: DiverseSiblingsSearchConfig):
        super().__init__(tgt_dict)
        self.diversity_rate = config.diversity_rate
        self.beam = BeamSearch(tgt_dict)

    def step(
        self,
        step: int,
        lprobs,
        scores,
        prev_output_tokens: Optional[Tensor] = None,
        original_batch_idxs: Optional[Tensor] = None,
    ):
        bsz, beam_size, vocab_size = lprobs.size()
        k = min(
            # Take the best 2 x beam_size predictions. We'll choose the first
            # beam_size of these which don't predict eos to continue with.
            beam_size * 2,
            lprobs.view(bsz, -1).size(1) - 1,  # -1 so we never select pad
        )
        s_list: List[Tensor]
        i_list: List[Tensor]
        s_list = [torch.empty(0).to(lprobs) for i in range(beam_size)]
        i_list = [torch.LongTensor().to(device=lprobs.device) for i in range(beam_size)]
        sibling_score = torch.arange(1, k + 1).to(lprobs) * self.diversity_rate

        if step == 0:
            return self.beam.step(step, lprobs, scores)
        lprobs.add_(scores[:, :, step - 1].unsqueeze(-1))

        # 1/ Calculate hypotheses for each beam
        for i in range(beam_size):
            torch.topk(lprobs[:, i, :].view(bsz, -1), k, out=(s_list[i], i_list[i]))
            i_list[i].fmod_(vocab_size)

            # 2/ Intra-sibling ordering by default from topk + 3/ Rewrite scores
            s_list[i].sub_(sibling_score)

        # 4/ Choose top K hypotheses
        indices = torch.stack(i_list, dim=1).view(bsz, -1)

        final_scores = torch.empty(0).to(lprobs)
        final_indices = torch.LongTensor().to(device=lprobs.device)
        final_beams = torch.LongTensor().to(device=lprobs.device)
        (final_scores, final_indices) = torch.topk(
            torch.stack(s_list, dim=1).view(bsz, -1),
            k,
        )

        final_beams = final_indices // k

        for i in range(bsz):
            final_indices[i] = indices[i][final_indices[i]]

        return final_scores, final_indices, final_beams



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, Union

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.data.postprocessor import BasePostProcessor, BasePostProcessorConfig
from dlk.utils.register import register


@cregister("postprocessor", "identity")
class IdentityPostProcessorConfig(BasePostProcessorConfig):
    """identity postprocessor"""

    pass


@register("postprocessor", "identity")
class IdentityPostProcessor(BasePostProcessor):
    """identity postprocessor"""

    def __init__(self, config: IdentityPostProcessorConfig):
        super(IdentityPostProcessor, self).__init__(config)

    def process(self, stage, outputs, origin_data) -> Dict:
        """do nothing except gather the loss"""
        if "loss" in outputs:
            return {self.loss_name_map(stage): torch.mean(outputs["loss"])}
        return {}



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Dict, List, Set

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import BaseModule
from dlk.utils.register import register


@cregister("decoder", "linear_crf")
class DecoderLinearCRFConfig(Base):
    """the linear_crf decoder module"""

    input_size = IntField(value=MISSING, minimum=1, help="the input size")
    output_size = IntField(value=MISSING, minimum=1, help="the output size")
    reduction = StrField(
        value="mean",
        options=["none", "sum", "mean", "token_mean"],
        help="the reduction method",
    )

    class InputMap:
        embedding = StrField(value="embedding", help="the embedding")
        label_ids = StrField(value="label_ids", help="the label ids")
        attention_mask = StrField(value="attention_mask", help="the attention mask")

    input_map = NestField(value=InputMap, help="the input map of the linear_crf module")

    class OutputMap:
        predict_seq_label = StrField(
            value="predict_seq_label", help="the predict seq label"
        )
        loss = StrField(value="loss", help="the loss")

    output_map = NestField(
        value=OutputMap, help="the output map of the linear_crf module"
    )

    submodule = SubModule(
        {
            "module@linear": {
                "input_size": "@$$.input_size @lambda x: x",
                "output_size": "@$$.output_size @lambda x: x",
            },
            "module@crf": {
                "reduction": "@$$.reduction @lambda x: x",
                "output_size": "@$$.output_size @lambda x: x",
            },
        }
    )


@register("decoder", "linear_crf")
class DecoderLinearCRF(BaseModule):
    """use torch.nn.Linear get the emission probability and fit to CRF"""

    def __init__(self, config: DecoderLinearCRFConfig):
        super(DecoderLinearCRF, self).__init__(config)

        self.linear = register.get("module", "linear")(config["@module@linear"])
        self.crf = register.get("module", "crf")(config["@module@crf"])
        self.config = config

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        self.linear.init_weight(method)
        self.crf.init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do predict, only get the predict labels

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        return self.predict_step(inputs)

    def predict_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do predict, only get the predict labels

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        logits = self.linear(inputs[self.config.input_map.embedding])
        inputs[self.config.output_map.predict_seq_label] = self.crf(
            logits, inputs[self.config.input_map.attention_mask]
        )
        return inputs

    def training_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do training step, get the crf loss

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        logits = self.linear(inputs[self.config.input_map.embedding])
        loss = self.crf.training_step(
            logits,
            inputs[self.config.input_map.label_ids],
            inputs[self.config.input_map.attention_mask],
        )
        inputs[self.config.output_map.loss] = loss
        return inputs

    def validation_step(
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """do validation step, get the crf loss and the predict labels

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        logits = self.linear(inputs[self.config.input_map.embedding])
        loss = self.crf.training_step(
            logits,
            inputs[self.config.input_map.label_ids],
            inputs[self.config.input_map.attention_mask],
        )
        inputs[self.config.output_map.loss] = loss
        inputs[self.config.output_map.predict_seq_label] = self.crf(
            logits, inputs[self.config.input_map.attention_mask]
        )
        return inputs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any, Callable, Dict, List, Union

import torch

from dlk.predict import Predict

logger = logging.getLogger(__name__)


class OnlinePredict(Predict):
    """OnlinePredict"""

    def __init__(self, config: Union[str, dict], checkpoint: str, update_config=None):
        super(OnlinePredict, self).__init__(config, checkpoint, update_config)
        datamodule, _ = self.get_datamodule(
            self.dlk_config, {}, world_size=self.trainer.world_size
        )
        self.online = True
        self.datamodule = datamodule

    def predict(self, data):
        """init the model, datamodule, manager then predict the predict_dataloader

        Args:
            data: the preprocessed data

        Returns:
            None

        """
        # get data
        dataloader = self.datamodule.online_dataloader(data)
        result = []
        with torch.no_grad():
            for i, batch in enumerate(dataloader):
                result.append(self.imodel.predict_step(batch, i))
        # start predict
        return self.imodel.postprocessor(
            stage="online",
            list_batch_outputs=result,
            origin_data=data,
            rt_config={},
            save_condition=False,
        )



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import Module


@cregister("module", "logits_gather")
class LogitsGatherConfig:
    """the logits_gather config"""

    prefix = StrField(value="gather_logits_", help="the prefix of the output name")
    logits_name_map = StrField(
        value="logits", help="the logits name map, gather which column"
    )
    gather_layer = DictField(
        value={},
        help="""
        the gather layer config, like
        "gather_layer": {
            "0": {
                "map": "3", # the 0th layer not do scale output to "gather_logits_3", "gather_logits_" is the output name prefix, the "3" is map name
                },
            "1": {
                "map": "4",
                }
            },
        default is empty dict
        """,
    )


@register("module", "logits_gather")
class LogitsGather(Module):
    """Gather the output logits decided by config"""

    def __init__(self, config: LogitsGatherConfig):
        super(LogitsGather, self).__init__()
        self.config = config
        self.layer_map: Dict[str, str] = {}
        self.prefix = self.config.prefix
        for layer, layer_config in self.config.gather_layer.items():
            self.layer_map[str(layer)] = str(layer_config["map"])

        if not self.layer_map:
            self.pass_gather = True
        else:
            self.pass_gather = False

    def forward(self, input: List[torch.Tensor]) -> Dict[str, torch.Tensor]:
        """gather the needed input to dict

        Args:
            batch: a mini batch inputs

        Returns:
            some elements to dict

        """
        result = torch.jit.annotate(Dict[str, torch.Tensor], {})

        if self.pass_gather:
            return result
        for layer, layer_suffix in self.layer_map.items():
            result[self.prefix + layer_suffix] = input[int(layer)]

        return result



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import random
import re
from typing import Dict, List

import numpy as np
import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import AdvMethod, AdvMethodConfig

logger = logging.getLogger(__name__)


@cregister("adv_method", "free_lb")
class FreeLBAdvMethodConfig(AdvMethodConfig):
    """
    ENHANCED ADVERSARIAL TRAINING FOR NATURAL LANGUAGE UNDERSTANDING
    https://arxiv.org/pdf/1909.11764.pdf
    """

    embedding_pattern = StrField(
        value="model.*embedding.*embedding",
        help="the pattern of embedding name that FreeLB effect on",
    )
    epsilon = FloatField(
        value=1.0,
        minimum=0.0,
        help="epsilon for FreeLB adversarial training",
    )
    alpha = FloatField(
        value=0.3,
        minimum=0.0,
        help="alpha for FreeLB adversarial training",
    )
    adv_k = IntField(value=3, minimum=1, help="FreeLB adversarial training times")


@register("adv_method", "free_lb")
class FreeLBAdvMethod(AdvMethod):
    """free_lb adversarial training method"""

    def __init__(self, model: nn.Module, config: FreeLBAdvMethodConfig):
        super().__init__(model, config)
        self.model = model
        self.config = config
        self.emb_backup = {}
        self.grad_backup = {}
        self.adv_para_name = set()
        for name, param in self.model.named_parameters():
            if param.requires_grad and re.findall(self.config.embedding_pattern, name):
                self.adv_para_name.add(name)
        logger.info(
            f"There are {len(self.adv_para_name)} paras will be adversarial training."
        )
        logger.info(f"They are {self.adv_para_name}.")

    def attack(self, is_first_attack=False):
        for name, param in self.model.named_parameters():
            if param.requires_grad and name in self.adv_para_name:
                if is_first_attack:
                    self.emb_backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm != 0 and not torch.isnan(norm):
                    r_at = self.config.alpha * param.grad / norm
                    param.data.add_(r_at)
                    param.data = self.project(name, param.data, self.config.epsilon)

    def project(self, param_name, param_data, epsilon):
        r = param_data - self.emb_backup[param_name]
        if torch.norm(r) > epsilon:
            r = epsilon * r / torch.norm(r)
        return self.emb_backup[param_name] + r

    def restore(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and name in self.adv_para_name:
                assert name in self.emb_backup
                param.data = self.emb_backup[name]
        self.emb_backup = {}

    def backup_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None:
                self.grad_backup[name] = param.grad.clone()

    def restore_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and name in self.grad_backup:
                param.grad = param.grad + self.grad_backup[name]

    def training_step(self, imodel, batch: Dict[str, torch.Tensor], batch_idx: int):
        """do training_step on a mini batch

        Args:
            imodel: imodel instance
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch

        Returns:
            the outputs

        """
        optimizer = imodel.optimizers()
        rt_config = {
            "current_step": imodel.global_step,
            "current_epoch": imodel.current_epoch,
            "total_steps": imodel.num_training_steps,
            "total_epochs": imodel.num_training_epochs,
        }
        optimizer.zero_grad()
        seed = random.randint(0, int(4e9))  # 4e9 < 2e32 - 1
        torch.manual_seed(seed)  # NOTE: should fix manual seed for every forward
        np.random.seed(seed)
        result = imodel.model.training_step(batch)
        loss, loss_log = imodel.calc_loss(result, batch, rt_config=rt_config)
        imodel.manual_backward(loss)

        self.backup_grad()
        for t in range(self.config.adv_k):
            self.attack(is_first_attack=(t == 0))
            if t == 0:
                optimizer.zero_grad()

            result = imodel.model.training_step(batch)
            loss, loss_log = imodel.calc_loss(result, batch, rt_config=rt_config)
            loss = loss / self.config.adv_k
            imodel.manual_backward(loss)
        self.restore_grad()
        self.restore()
        optimizer.step()

        schedule = imodel.lr_schedulers()
        schedule.step()
        return loss, loss_log
